from __future__ import annotations

from configparser import ConfigParser
from pathlib import Path

from grascii.config import get_config_file_path

_CONFIG = ConfigParser()
_CONFIG.read_file(Path(__file__).with_name("defaults.conf").open())
_CONFIG.read(get_config_file_path())

DEFAULTS = ConfigParser()
DEFAULTS.read_file(Path(__file__).with_name("defaults.conf").open())

SEARCH = _CONFIG["Search"]



from __future__ import annotations

import logging
from pathlib import Path

from grascii import DictionaryBuilder, DictionaryOutputOptions


def check_submodule():
    if not Path("dictionaries/builtins").exists():
        logging.critical(
            "dictionaries/builtins does not exist. "
            + "Has the submodule been checked out?\n"
            + "Try: git submodule update --init"
        )
        raise RuntimeError("Missing dictionary build requirements")


def build_dictionaries():
    check_submodule()
    builder = DictionaryBuilder()
    builder.build(
        infiles=Path("dictionaries/builtins/preanniversary").glob("*.txt"),
        output=DictionaryOutputOptions(
            output_dir=Path("grascii/dictionary/preanniversary"),
            package=True,
        ),
    )


if __name__ == "__main__":
    build_dictionaries()



from __future__ import annotations

import re
from functools import lru_cache, reduce
from typing import Iterator, List, Union

from lark import Lark, Token, Transformer, Tree, UnexpectedInput

from grascii import grammar
from grascii.lark_ambig_tools import Disambiguator

Interpretation = List[Union[str, List[str]]]


def interpretation_to_string(interpretation: Interpretation) -> str:
    """Generate a string representation of an Interpretation.

    :param interpretation: An Interpretation to generate a string for.
    :returns: A string representation of an Interpretation
    """

    def reducer(builder, token):
        if isinstance(token, list):
            builder += token
        else:
            if (
                builder
                and builder[-1] != grammar.DISJOINER
                and token != grammar.DISJOINER
                and builder[-1] != grammar.BOUNDARY
                and token != grammar.BOUNDARY
            ):
                builder.append(grammar.BOUNDARY)
            if token != grammar.BOUNDARY:
                builder.append(token)
        return builder

    return "".join(reduce(reducer, interpretation, []))


class GrasciiFlattener(Transformer):

    """This is a Lark Transformer that converts a parsed Grascii string
    into an ``Interpretation``. An ``Interpretation`` is a list of terminals and
    annotation lists. Each terminal is its own element in the interpretation,
    but sequences of annotation terminals are grouped into their own sublist.
    """

    def __init__(
        self, preserve_boundaries: bool = False, start_rule: str = "start"
    ) -> None:
        super().__init__()
        setattr(self, start_rule, self.start)
        self.preserve_boundaries = preserve_boundaries

    def start(self, children) -> Interpretation:
        """Returns the final result of the transformation.

        :meta private:
        """

        result: Interpretation = []
        for child in children:
            for token in child:
                if token in grammar.ANNOTATION_CHARACTERS:
                    # pack annotation terminals into sublist
                    if isinstance(result[-1], list):
                        # add to previous annotation list
                        result[-1].append(token)
                    else:
                        # start new annotation list
                        result.append([token])
                elif token != grammar.BOUNDARY or self.preserve_boundaries:
                    result.append(token)
        return result

    def __default__(self, data, children, meta) -> List[str]:
        """The default visitor function for nodes in the parse tree.
        Returns a flat list of strings."""

        result: List[str] = []
        for child in children:
            if isinstance(child, Token):
                result.append(child)
            else:
                # flatten iterable
                for token in child:
                    result.append(token)
        return result


@lru_cache(maxsize=1)
def get_grascii_regex_str() -> str:
    """Get a string that can be compiled into a regular expression that matches
    Grascii strings.
    """
    parser = Lark.open_from_package("grascii.grammars", "grascii_regex.lark")
    return parser.get_terminal("GRASCII").pattern.value


class GrasciiValidator:
    """Validates Grascii strings.

    :param ignore_case: Whether to ignore the case of the Grascii string. If
        ``False``, the Grascii string must be uppercase.
    :type ignore_case: bool
    """

    def __init__(self, ignore_case: bool = False) -> None:
        self._regex = re.compile(get_grascii_regex_str(), re.I if ignore_case else 0)

    def validate(self, grascii: str) -> bool:
        """Check whether the given string is valid Grascii.

        :param grascii: A string to check
        :returns: bool
        """
        return bool(self._regex.fullmatch(grascii))


class InvalidGrascii(Exception):
    """Exception thrown by the ``GrasciiParser`` when provided an invalid string."""

    def __init__(self, grascii: str, unexpected_input: UnexpectedInput) -> None:
        self.grascii = grascii
        self.unexpected_input = unexpected_input
        self.context = unexpected_input.get_context(grascii)


class GrasciiParser:
    """Parses and interprets Grascii strings."""

    def __init__(self) -> None:
        self._parser: Lark = Lark.open_from_package(
            "grascii.grammars", "grascii.lark", parser="earley", ambiguity="explicit"
        )

    def parse(self, grascii: str) -> Tree:
        """Parse the given string into a ``Tree``.

        :param grascii: A Grascii string to parse.
        :returns: An ambiguous parse tree of the Grascii string.
        """
        try:
            return self._parser.parse(grascii)
        except UnexpectedInput as ui:
            raise InvalidGrascii(grascii, ui)

    def interpret(
        self, grascii: str, preserve_boundaries: bool = False
    ) -> Iterator[Interpretation]:
        """Interpret a Grascii string.

        :param grascii: A Grascii string to interpret
        :param preserve_boundaries: When ``False``, boundaries in the string ('-')
            are not included in the resulting interpretations.
        :returns: An iterator over interpretations.
        """
        tree = self.parse(grascii)
        trees = Disambiguator().visit(tree)
        flattener = GrasciiFlattener(preserve_boundaries)
        return map(flattener.transform, trees)



from __future__ import annotations

import argparse
import sys
from functools import lru_cache
from typing import Set

from lark import Lark, Token, Transformer, UnexpectedInput
from lark.visitors import VisitError, v_args

from grascii.lark_ambig_tools import Disambiguator
from grascii.parser import GrasciiFlattener, interpretation_to_string
from grascii.searchers import GrasciiSearcher

description = "Decipher a shorthand phrase."


def build_argparser(argparser: argparse.ArgumentParser) -> None:
    argparser.add_argument("phrase", action="store", help="The phrase to decipher")
    argparser.add_argument(
        "-a",
        "--aggressive",
        action="store_true",
        default=False,
        help="perform a more intense dephrasing",
    )
    argparser.add_argument(
        "--ignore-limit",
        action="store_true",
        default=False,
        help="ignore the 8-character phrase limit",
    )


class NoWordFound(Exception):
    pass


class StripNameSpace(Transformer):
    def __init__(self, namespace):
        self.namespace = namespace + "__"

    def __default__(self, data, children, meta):
        if data.startswith(self.namespace):
            data = data[len(self.namespace) :]
        return super().__default__(data, children, meta)


class PhraseFlattener(Transformer):
    optionals = {
        "opt_to": "TO",
        "opt_in": "IN",
        "opt_of": "OF",
        "opt_your": "YOUR",
    }

    _grascii_flattener = GrasciiFlattener(start_rule="word")

    _grascii_searcher = GrasciiSearcher()

    def __init__(self):
        for key, value in self.optionals.items():
            setattr(self, key, self.make_opt(value))

    def make_opt(self, name):
        def opt(children):
            if len(children):
                return self.__default__(None, children, None)
            return [self.create_token("[" + name + "]")]

        return opt

    def start(self, children):
        result = []
        for child in children:
            for token in child:
                result.append(token)
        return result

    def create_token(self, name: str, value: str = ""):
        return Token(name, value)

    def short_to(self, children):
        return [self.create_token("TO")]

    @lru_cache(maxsize=32)
    def _search_grascii(self, grascii_str):
        results = self._grascii_searcher.sorted_search(grascii=grascii_str)
        if not results:
            raise NoWordFound()
        words = (result.entry.translation.upper() for result in results)
        string = "(" + "|".join(words) + ")"
        return self.create_token(string)

    @v_args(tree=True)
    def word(self, tree):
        interp = self._grascii_flattener.transform(tree)
        return self._search_grascii(interpretation_to_string(interp))

    def __default__(self, data, children, meta):
        result = []
        for child in children:
            if isinstance(child, Token):
                # strip namespace
                i = child.type.rfind("_")
                new_type = child.type[i + 1 :]
                result.append(self.create_token(new_type, child.value))
            else:
                for token in child:
                    result.append(token)
        return result


def dephrase(**kwargs) -> Set[str]:
    aggressive = kwargs.get("aggressive", False)
    grammar_name = "phrases_extended.lark" if aggressive else "phrases.lark"
    parser = Lark.open_from_package(
        "grascii.grammars",
        grammar_name,
        parser="earley",
        ambiguity="explicit",
        lexer="dynamic_complete",
    )
    trans = PhraseFlattener()
    if aggressive:
        trans = StripNameSpace("phrases") * trans
    parses: Set[str] = set()
    try:
        tree = parser.parse(kwargs["phrase"].upper())
    except UnexpectedInput:
        return parses

    trees = Disambiguator().visit(tree)
    for t in trees:
        try:
            tokens = (token.type for token in trans.transform(t))
        except VisitError as e:
            if isinstance(e.orig_exc, NoWordFound):
                continue
            raise e  # no cov
        else:
            parses.add(" ".join(tokens))
    return parses


def cli_dephrase(args: argparse.Namespace) -> None:
    if len(args.phrase) > 8 and args.aggressive and not args.ignore_limit:
        print(
            "Phrases more than 8 characters in length may take",
            "an excessively long time to process and produce many",
            "irrelevant results.",
        )
        print("To ignore this warning use '--ignore-limit'.")
        return
    results = dephrase(**{k: v for k, v in vars(args).items() if v is not None})
    if results:
        for result in results:
            print(result)
    else:
        print("No results")
        if not args.aggressive:
            print("You may try again with --aggressive to consider more possibilities.")


def main() -> None:
    argparser = argparse.ArgumentParser(description)
    build_argparser(argparser)
    args = argparser.parse_args(sys.argv[1:])
    cli_dephrase(args)


if __name__ == "__main__":
    main()



from __future__ import annotations

import unittest

import pytest

try:
    import pexpect
except ImportError:
    pexpect = None

SUPPORTS_INTERACTIVE = False
try:
    import grascii.interactive  # noqa: F401

    SUPPORTS_INTERACTIVE = True
except ImportError:
    pass

DOWN = "j"
UP = "k"


@pytest.mark.slow
@unittest.skipIf(pexpect is None, "pexpect is not installed")
@unittest.skipUnless(SUPPORTS_INTERACTIVE, "interactive extra is not installed")
class InteractiveTester(unittest.TestCase):
    def setUp(self):
        self.c = pexpect.spawn(
            "grascii search --interactive --dictionary :preanniversary"
        )

    def tearDown(self):
        self.c.terminate(force=True)

    def expect(self, patterns):
        return self.c.expect([pexpect.TIMEOUT] + patterns, timeout=5)

    def new_search(self):
        self.c.sendline()

    def navigate_to_settings(self):
        self.c.send(UP)
        self.c.send(UP)
        self.c.sendline()

    def cancel(self):
        self.c.sendcontrol("c")

    def assert_on_main_menu(self):
        self.assertGreater(self.expect(["What would you like to do?"]), 0)

    def assert_on_settings(self):
        self.assertGreater(self.expect(["Search Settings"]), 0)

    def assert_modify_search_enabled(self):
        self.assertGreater(self.expect(["(?<!- )Modify Search"]), 0)

    def assert_modify_search_disabled(self):
        self.assertGreater(self.expect(["- Modify Search"]), 0)


class TestMenus(InteractiveTester):
    def test_main_menu(self):
        self.assertGreater(self.expect(["What would you like to do?"]), 0)
        self.assertGreater(self.expect(["New Search"]), 0)
        self.assertGreater(self.expect(["- Modify Search"]), 0)
        self.assertGreater(self.expect(["Edit Settings"]), 0)
        self.assertGreater(self.expect(["Exit"]), 0)

    def test_exit(self):
        self.c.sendline(UP)
        self.c.sendline()
        self.assertGreater(self.expect([pexpect.EOF]), 0)
        self.assertFalse(self.c.isalive())

    def test_new_search(self):
        self.c.sendline()
        self.assertGreater(self.expect(["Enter Search:"]), 0)

    def test_settings(self):
        self.navigate_to_settings()
        self.assertGreater(self.expect(["Search Settings"]), 0)
        self.assertGreater(self.expect(["Uncertainty"]), 0)
        self.assertGreater(self.expect(["Search Mode"]), 0)
        self.assertGreater(self.expect(["Annotation Mode"]), 0)
        self.assertGreater(self.expect(["Aspirate Mode"]), 0)
        self.assertGreater(self.expect(["Disjoiner Mode"]), 0)
        self.assertGreater(self.expect(["Fix First"]), 0)
        self.assertGreater(self.expect(["Dictionaries"]), 0)
        self.assertGreater(self.expect(["Back"]), 0)

    def test_uncertainty(self):
        self.navigate_to_settings()
        self.c.sendline()
        self.assertGreater(self.expect(["Uncertainty"]), 0)
        self.assertGreater(self.expect(["0"]), 0)
        self.assertGreater(self.expect(["1"]), 0)
        self.assertGreater(self.expect(["2"]), 0)

    def test_search_mode(self):
        self.navigate_to_settings()
        self.c.send(DOWN)
        self.c.sendline()
        self.assertGreater(self.expect(["Search Mode"]), 0)
        self.assertGreater(self.expect(["match"]), 0)
        self.assertGreater(self.expect(["start"]), 0)
        self.assertGreater(self.expect(["contain"]), 0)

    def test_annotation_mode(self):
        self.navigate_to_settings()
        for i in range(2):
            self.c.send(DOWN)
        self.c.sendline()
        self.assertGreater(self.expect(["Annotation Mode"]), 0)
        self.assertGreater(self.expect(["discard"]), 0)
        self.assertGreater(self.expect(["retain"]), 0)
        self.assertGreater(self.expect(["strict"]), 0)

    def test_aspirate_mode(self):
        self.navigate_to_settings()
        for i in range(3):
            self.c.send(DOWN)
        self.c.sendline()
        self.assertGreater(self.expect(["Aspirate Mode"]), 0)
        self.assertGreater(self.expect(["discard"]), 0)
        self.assertGreater(self.expect(["retain"]), 0)
        self.assertGreater(self.expect(["strict"]), 0)

    def test_disjoiner_mode(self):
        self.navigate_to_settings()
        for i in range(4):
            self.c.send(DOWN)
        self.c.sendline()
        self.assertGreater(self.expect(["Disjoiner Mode"]), 0)
        self.assertGreater(self.expect(["discard"]), 0)
        self.assertGreater(self.expect(["retain"]), 0)
        self.assertGreater(self.expect(["strict"]), 0)

    def test_fix_first(self):
        self.navigate_to_settings()
        for i in range(5):
            self.c.send(DOWN)
        self.c.sendline()
        self.assertGreater(self.expect(["Fix First"]), 0)
        self.assertGreater(self.expect(["True"]), 0)
        self.assertGreater(self.expect(["False"]), 0)

    def test_dictionaries(self):
        self.navigate_to_settings()
        for i in range(6):
            self.c.send(DOWN)
        self.c.sendline()
        self.assertGreater(self.expect(["Choose Dictionaries"]), 0)
        self.assertGreater(self.expect([":preanniversary"]), 0)

    def test_settings_back(self):
        self.navigate_to_settings()
        self.assert_on_settings()
        self.c.send(UP)
        self.c.sendline()
        self.assert_on_main_menu()


class TestSearch(InteractiveTester):
    def test_no_search_input(self):
        self.new_search()
        for i in range(4):
            self.c.sendline()
            self.assertGreater(self.expect(["Enter Search:"]), 0)

    def test_search(self):
        self.new_search()
        self.c.sendline("a")
        self.assertGreater(self.expect(["Search Results"]), 0)
        self.assertGreater(self.expect(["Next"]), 0)
        self.assertGreater(self.expect(["Display All"]), 0)
        self.assertGreater(self.expect(["End Search"]), 0)
        i = 2
        while i == 2:
            self.c.sendline()
            i = self.expect(["Results: \\d+", "Next"])
        self.assertEqual(i, 1)
        self.assert_on_main_menu()

    def test_display_all(self):
        self.new_search()
        self.c.sendline("a")
        self.assertGreater(self.expect(["Search Results"]), 0)
        self.c.send(DOWN)
        self.c.sendline()
        self.assertGreater(self.expect(["Results: \\d+"]), 0)
        self.assert_on_main_menu()

    def test_end_search_immediately(self):
        self.new_search()
        self.c.sendline("a")
        self.assertGreater(self.expect(["Search Results"]), 0)
        self.c.send(DOWN)
        self.c.send(DOWN)
        self.c.sendline()
        self.assertGreater(self.expect(["Results: 0"]), 0)
        self.assert_on_main_menu()

    def test_end_search_later(self):
        self.new_search()
        self.c.sendline("a")
        self.assertGreater(self.expect(["Search Results"]), 0)
        self.c.sendline()
        self.assertGreater(self.expect(["Search Results"]), 0)
        self.c.sendline()
        self.assertGreater(self.expect(["Search Results"]), 0)
        self.c.send(DOWN)
        self.c.send(DOWN)
        self.c.sendline()
        self.assertGreater(self.expect(["Results: 2"]), 0)
        self.assert_on_main_menu()

    def test_interpretation(self):
        self.new_search()
        self.c.sendline("tn")
        self.assertGreater(self.expect(["all"]), 0)
        self.assertGreater(self.expect(["TN"]), 0)
        self.assertGreater(self.expect(["T-N"]), 0)
        self.c.sendline()
        self.assertGreater(self.expect(["Search Results"]), 0)

    def test_modify_search(self):
        self.new_search()
        self.assertGreater(self.expect(["Enter Search:"]), 0)
        self.c.sendline("a")
        self.c.send(DOWN)
        self.c.sendline()
        self.assert_on_main_menu()
        self.assert_modify_search_enabled()
        self.c.send(DOWN)
        self.c.sendline()
        self.assertGreater(self.expect(["Enter Search:"]), 0)
        self.c.sendline("b")
        self.assertGreater(self.expect(["Search Results"]), 0)
        self.c.sendline()
        self.assertGreater(self.expect(["AB"]), 0)

    def test_invalid_grascii(self):
        self.new_search()
        self.assertGreater(self.expect(["Enter Search:"]), 0)
        self.c.sendline("rac")
        self.assertGreater(self.expect(["Invalid Grascii"]), 0)


class TestCancel(InteractiveTester):
    def test_main_menu(self):
        self.assert_on_main_menu()
        self.cancel()
        self.assertGreater(self.expect([pexpect.EOF]), 0)
        self.assertFalse(self.c.isalive())

    def test_new_search(self):
        self.new_search()
        self.assertGreater(self.expect(["Enter Search:"]), 0)
        self.cancel()
        self.assert_on_main_menu()
        self.assert_modify_search_disabled()

    def test_input_new_search(self):
        self.new_search()
        self.assertGreater(self.expect(["Enter Search:"]), 0)
        self.c.send("a")
        self.cancel()
        self.assert_on_main_menu()
        self.assert_modify_search_disabled()

    def test_search_results(self):
        self.new_search()
        self.assertGreater(self.expect(["Enter Search:"]), 0)
        self.c.send("a")
        self.c.sendline()
        self.cancel()
        self.assertGreater(self.expect(["Results: 0"]), 0)
        self.assert_on_main_menu()
        self.assert_modify_search_enabled()

    def test_search_results_later(self):
        self.new_search()
        self.assertGreater(self.expect(["Enter Search:"]), 0)
        self.c.send("a")
        self.c.sendline()
        self.c.sendline()
        self.cancel()
        self.assertGreater(self.expect(["Results: 1"]), 0)
        self.assert_on_main_menu()
        self.assert_modify_search_enabled()

    def test_interpretation(self):
        self.new_search()
        self.assertGreater(self.expect(["Enter Search:"]), 0)
        self.c.sendline("tn")
        self.cancel()
        self.assert_on_main_menu()
        self.assert_modify_search_enabled()

    def test_settings(self):
        self.navigate_to_settings()
        self.assert_on_settings()
        self.cancel()
        self.assert_on_main_menu()

    def test_uncertainty(self):
        self.navigate_to_settings()
        self.assert_on_settings()
        self.c.sendline()
        self.cancel()
        self.assert_on_settings()

    def test_search_mode(self):
        self.navigate_to_settings()
        self.assert_on_settings()
        self.c.send(DOWN)
        self.c.sendline()
        self.cancel()
        self.assert_on_settings()

    def test_annotation_mode(self):
        self.navigate_to_settings()
        self.assert_on_settings()
        for i in range(2):
            self.c.send(DOWN)
        self.c.sendline()
        self.cancel()
        self.assert_on_settings()

    def test_aspirate_mode(self):
        self.navigate_to_settings()
        self.assert_on_settings()
        for i in range(3):
            self.c.send(DOWN)
        self.c.sendline()
        self.cancel()
        self.assert_on_settings()

    def test_disjoiner_mode(self):
        self.navigate_to_settings()
        self.assert_on_settings()
        for i in range(4):
            self.c.send(DOWN)
        self.c.sendline()
        self.cancel()
        self.assert_on_settings()

    def test_fix_first(self):
        self.navigate_to_settings()
        self.assert_on_settings()
        for i in range(5):
            self.c.send(DOWN)
        self.c.sendline()
        self.cancel()
        self.assert_on_settings()

    def test_dictionaries(self):
        self.navigate_to_settings()
        self.assert_on_settings()
        for i in range(6):
            self.c.send(DOWN)
        self.c.sendline()
        self.cancel()
        self.assert_on_settings()



from __future__ import annotations

from pathlib import Path

from lark import Lark, Token, Transformer, Tree
from lark.reconstruct import Reconstructor


class GrasciiProcessor(Transformer):
    def __init__(self, parser):
        super().__init__(visit_tokens=True)
        self.parser = parser

    def start(self, children):
        # insert a new start rule
        new_start = self.parser.parse("start: GRASCII", start="rule")
        return Tree(
            "start", [new_start, Token("_NL", "\n"), Token("_NL", "\n")] + children
        )

    def rule(self, children):
        if children[0] == "_STRING":
            # insert a nonrecursive definition of _string
            new_string = self.parser.parse(
                "_STRING: _LETTER (ASPIRATE? (BOUNDARY | INTERSECTION)? _LETTER)*",
                start="token",
            )
            return new_string
        return Tree("token", children)

    def rule_params(self, children):
        return Tree("token_params", children)

    def RULE(self, token):
        if token == "start":
            # rename the start rule to a token
            return Token("TOKEN", "GRASCII")
        # change all rules to tokens
        return Token("TOKEN", token.upper())


def postproc(strings):
    """Insert necessary spaces back into the lark grammar"""
    prev = ""
    for item in strings:
        if getattr(prev, "type", "") == "OP" and getattr(item, "type", "") in {
            "RULE",
            "TOKEN",
        }:
            yield " "
        yield item
        prev = item


def create_grascii_regex():
    parser = Lark.open_from_package(
        "lark",
        "grammars/lark.lark",
        maybe_placeholders=False,
        parser="lalr",
        start=["start", "token", "rule"],
        keep_all_tokens=True,
    )

    tree = parser.parse(
        Path("grascii/grammars/grascii.lark").read_text(), start="start"
    )
    tree = GrasciiProcessor(parser).transform(tree)

    reconstructor = Reconstructor(parser)
    grammar = reconstructor.reconstruct(tree, postproc)
    Path("grascii/grammars/grascii_regex.lark").write_text(grammar)
    print("Created grascii/grammars/grascii_regex.lark")
    print(grammar)


if __name__ == "__main__":
    create_grascii_regex()



from __future__ import annotations

from pathlib import Path

import pytest

from grascii.config import (
    CONF_FILE_NAME,
    config_exists,
    create_config,
    delete_config,
    get_config_file_path,
    get_default_config,
)


@pytest.fixture
def tmp_config_path(tmp_path, monkeypatch):
    monkeypatch.setattr("grascii.config.CONF_DIRECTORY", tmp_path)
    return tmp_path


def test_get_config_file_path(tmp_config_path):
    assert get_config_file_path() == tmp_config_path / CONF_FILE_NAME


def test_get_default_config():
    assert Path("grascii/defaults.conf").read_text() == get_default_config()


def test_create_and_delete_config(tmp_config_path):
    assert not config_exists()
    create_config()
    assert config_exists()
    assert Path(tmp_config_path, CONF_FILE_NAME).exists()
    assert Path(get_config_file_path()).read_text() == get_default_config()
    delete_config()
    assert not Path(tmp_config_path, CONF_FILE_NAME).exists()
    assert not config_exists()



from __future__ import annotations

import argparse
import sys
from typing import Collection

from grascii.dictionary.common import INSTALLATION_DIR, get_dictionary_installed_name

description = "List built-in and installed dictionaries."


def build_argparser(argparser: argparse.ArgumentParser) -> None:
    return


def get_built_ins() -> Collection[str]:
    """Get a collection of the installed names of all built-in dictionaries.

    :returns: A collection of built-in dictionary names.
    """
    if sys.version_info >= (3, 9):
        from importlib.resources import files
    else:
        from importlib_resources import files
    entries = files("grascii.dictionary").iterdir()
    built_ins = filter(lambda e: e.is_dir() and e.name[0] != "_", entries)
    return list(map(lambda e: get_dictionary_installed_name(e.name), built_ins))


def get_installed() -> Collection[str]:
    """Get a collection of the installed names of all user-installed dictionaries.

    :returns: A collection of installed dictionary names.
    """
    if INSTALLATION_DIR.exists():
        files = filter(lambda f: f.is_dir(), INSTALLATION_DIR.iterdir())
        installed = map(lambda f: f.name, files)
        return list(map(get_dictionary_installed_name, installed))
    return []


def cli_list(args: argparse.Namespace) -> None:
    print("Built-in Dictionaries:")
    for built_in in get_built_ins():
        print(built_in)
    print()

    print("Installed Dictionaries:")
    for installed in get_installed():
        print(installed)


def main() -> None:
    """Run the list command using arguments from sys.argv."""

    argparser = argparse.ArgumentParser(description)
    build_argparser(argparser)
    args = argparser.parse_args(sys.argv[1:])
    cli_list(args)


if __name__ == "__main__":
    main()



from __future__ import annotations

import unittest

from grascii.metrics import gsequence_distance, interpretation_to_gsequence


class TestStandard(unittest.TestCase):
    def test_order(self):
        base = interpretation_to_gsequence(["A", "B"])
        aspirate = interpretation_to_gsequence(["'", "A", "B"])
        medium_sound = interpretation_to_gsequence(["A", ["."], "B"])
        long_sound = interpretation_to_gsequence(["A", [","], "B"])
        wunderbar = interpretation_to_gsequence(["A", ["_"], "B"])
        loop = interpretation_to_gsequence(["A", ["|"], "B"])
        reverse = interpretation_to_gsequence(["A", ["~"], "B"])
        disjoiner = interpretation_to_gsequence(["A", "^", "B"])
        uncertainty_one = interpretation_to_gsequence(["A", "P"])
        uncertainty_two = interpretation_to_gsequence(["A", "S"])
        insert_one = interpretation_to_gsequence(["A", "K", "B"])
        insert_two = interpretation_to_gsequence(["A", "K", "B", "D"])
        self.assertEqual(gsequence_distance(base, base), 0)
        self.assertGreater(gsequence_distance(aspirate, base), 0)
        self.assertEqual(
            gsequence_distance(medium_sound, base), gsequence_distance(long_sound, base)
        )
        self.assertGreaterEqual(
            gsequence_distance(medium_sound, base), gsequence_distance(aspirate, base)
        )
        self.assertGreaterEqual(
            gsequence_distance(wunderbar, base), gsequence_distance(medium_sound, base)
        )
        self.assertGreaterEqual(
            gsequence_distance(loop, base), gsequence_distance(wunderbar, base)
        )
        self.assertGreaterEqual(
            gsequence_distance(reverse, base), gsequence_distance(loop, base)
        )
        self.assertGreaterEqual(
            gsequence_distance(disjoiner, base), gsequence_distance(reverse, base)
        )
        self.assertGreaterEqual(
            gsequence_distance(uncertainty_one, base),
            gsequence_distance(disjoiner, base),
        )
        self.assertGreater(
            gsequence_distance(uncertainty_two, base),
            gsequence_distance(uncertainty_one, base),
        )
        self.assertGreaterEqual(
            gsequence_distance(insert_one, base), gsequence_distance(disjoiner, base)
        )
        self.assertGreater(
            gsequence_distance(insert_two, base), gsequence_distance(insert_one, base)
        )

    def test_uncertainty(self):
        base = interpretation_to_gsequence(["A", "S"])
        az = interpretation_to_gsequence(["A", "Z"])
        ap = interpretation_to_gsequence(["A", "P"])
        ab = interpretation_to_gsequence(["A", "B"])
        af = interpretation_to_gsequence(["A", "F"])
        av = interpretation_to_gsequence(["A", "V"])
        self.assertEqual(gsequence_distance(base, base), 0)
        self.assertEqual(gsequence_distance(az, base), 0)
        self.assertGreater(gsequence_distance(ap, base), 0)
        self.assertEqual(gsequence_distance(ap, base), gsequence_distance(af, base))
        self.assertEqual(gsequence_distance(ab, base), gsequence_distance(av, base))
        self.assertGreater(gsequence_distance(av, base), gsequence_distance(af, base))

    def test_equality(self):
        aspirate = interpretation_to_gsequence(["'", "A", "B"])
        medium_sound = interpretation_to_gsequence(["A", ["."], "B"])
        long_sound = interpretation_to_gsequence(["A", [","], "B"])
        wunderbar = interpretation_to_gsequence(["A", ["_"], "B"])
        loop = interpretation_to_gsequence(["A", ["|"], "B"])
        reverse = interpretation_to_gsequence(["A", ["~"], "B"])
        disjoiner = interpretation_to_gsequence(["A", "^", "B"])
        uncertainty_one = interpretation_to_gsequence(["A", "P"])
        uncertainty_two = interpretation_to_gsequence(["A", "S"])
        self.assertEqual(gsequence_distance(aspirate, aspirate), 0)
        self.assertEqual(gsequence_distance(medium_sound, medium_sound), 0)
        self.assertEqual(gsequence_distance(long_sound, long_sound), 0)
        self.assertEqual(gsequence_distance(wunderbar, wunderbar), 0)
        self.assertEqual(gsequence_distance(loop, loop), 0)
        self.assertEqual(gsequence_distance(reverse, reverse), 0)
        self.assertEqual(gsequence_distance(disjoiner, disjoiner), 0)
        self.assertEqual(gsequence_distance(uncertainty_one, uncertainty_one), 0)
        self.assertEqual(gsequence_distance(uncertainty_two, uncertainty_two), 0)

    def test_symmetry(self):
        aspirate = interpretation_to_gsequence(["'", "A", "B"])
        medium_sound = interpretation_to_gsequence(["A", ["."], "B"])
        long_sound = interpretation_to_gsequence(["A", [","], "B"])
        wunderbar = interpretation_to_gsequence(["A", ["_"], "B"])
        loop = interpretation_to_gsequence(["A", ["|"], "B"])
        reverse = interpretation_to_gsequence(["A", ["~"], "B"])
        disjoiner = interpretation_to_gsequence(["A", "^", "B"])
        uncertainty_one = interpretation_to_gsequence(["A", "P"])
        uncertainty_two = interpretation_to_gsequence(["A", "S"])
        self.assertEqual(
            gsequence_distance(aspirate, medium_sound),
            gsequence_distance(medium_sound, aspirate),
        )
        self.assertEqual(
            gsequence_distance(long_sound, medium_sound),
            gsequence_distance(medium_sound, long_sound),
        )
        self.assertEqual(
            gsequence_distance(long_sound, wunderbar),
            gsequence_distance(wunderbar, long_sound),
        )
        self.assertEqual(
            gsequence_distance(loop, wunderbar), gsequence_distance(wunderbar, loop)
        )
        self.assertEqual(
            gsequence_distance(loop, uncertainty_one),
            gsequence_distance(uncertainty_one, loop),
        )
        self.assertEqual(
            gsequence_distance(reverse, uncertainty_one),
            gsequence_distance(uncertainty_one, reverse),
        )
        self.assertEqual(
            gsequence_distance(reverse, uncertainty_two),
            gsequence_distance(uncertainty_two, reverse),
        )
        self.assertEqual(
            gsequence_distance(disjoiner, uncertainty_two),
            gsequence_distance(uncertainty_two, disjoiner),
        )

    def test_triangle_inequality(self):
        base = interpretation_to_gsequence(["A", "B"])
        aspirate = interpretation_to_gsequence(["'", "A", "B"])
        medium_sound = interpretation_to_gsequence(["A", ["."], "B"])
        long_sound = interpretation_to_gsequence(["A", [","], "B"])
        wunderbar = interpretation_to_gsequence(["A", ["_"], "B"])
        loop = interpretation_to_gsequence(["A", ["|"], "B"])
        reverse = interpretation_to_gsequence(["A", ["~"], "B"])
        disjoiner = interpretation_to_gsequence(["A", "^", "B"])
        uncertainty_one = interpretation_to_gsequence(["A", "P"])
        uncertainty_two = interpretation_to_gsequence(["A", "S"])
        self.assertGreaterEqual(
            gsequence_distance(base, wunderbar)
            + gsequence_distance(wunderbar, uncertainty_one),
            gsequence_distance(base, uncertainty_one),
        )
        self.assertGreaterEqual(
            gsequence_distance(medium_sound, disjoiner)
            + gsequence_distance(disjoiner, aspirate),
            gsequence_distance(medium_sound, aspirate),
        )
        self.assertGreaterEqual(
            gsequence_distance(loop, reverse) + gsequence_distance(reverse, long_sound),
            gsequence_distance(loop, long_sound),
        )
        self.assertGreaterEqual(
            gsequence_distance(disjoiner, base)
            + gsequence_distance(base, uncertainty_two),
            gsequence_distance(disjoiner, uncertainty_two),
        )

    def test_direction(self):
        base = interpretation_to_gsequence(["A", "S"])
        left = interpretation_to_gsequence(["A", "S", ["("]])
        right = interpretation_to_gsequence(["A", "S", [")"]])
        self.assertGreater(gsequence_distance(left, base), 0)
        self.assertGreater(gsequence_distance(right, base), 0)
        self.assertEqual(
            gsequence_distance(left, base), gsequence_distance(right, base)
        )



"""
Contains RegexBuilder for generating regular expression to use in grascii
searches.
"""

from __future__ import annotations

import re
from enum import Enum
from typing import Iterable, List, Pattern, Set, Tuple

from grascii import grammar
from grascii.parser import Interpretation
from grascii.similarities import get_similar


class SearchMode(Enum):
    """An enum representing different search modes."""

    MATCH = "match"
    START = "start"
    CONTAIN = "contain"


class Strictness(Enum):
    """An enum representing different levels of strictness for handling
    annotations, aspirates, and disjoiners in grascii strings."""

    LOW = "discard"
    MEDIUM = "retain"
    HIGH = "strict"


class RegexBuilder:
    """A class used for generating regular expressions used to search from
    a grascii string based on grascii search options.

    :param uncertainty: The uncertainty of the grascii string.
    :param search_mode: The search mode to use.
    :param annotation_mode: How to handle annotations in the search.
    :param aspirate_mode: How to handle annotations in the search.
    :param disjoiner_mode: How to handle annotations in the search.
    :param fix_first: Apply an uncertainty of 0 to the first token.
    :type uncertainty: int: 0, 1, or 2
    :type search_mode: str: one of regen.SearchMode values
    :type annotation_mode: one of regen.Strictness values
    :type aspirate_mode: one of regen.Strictness values
    :type disjoiner_mode: one of regen.Strictness values
    :type fix_first: bool
    """

    def __init__(self, **kwargs):
        self.uncertainty = kwargs.get("uncertainty", 0)
        self.search_mode = kwargs.get("search_mode", SearchMode.MATCH)
        self.fix_first = kwargs.get("fix_first", False)
        self.annotation_mode = kwargs.get("annotation_mode", Strictness.LOW)
        self.aspirate_mode = kwargs.get("aspirate_mode", Strictness.LOW)
        self.disjoiner_mode = kwargs.get("disjoiner_mode", Strictness.HIGH)

    def make_annotation_regex(self, stroke: str, annotations: Iterable[str]) -> str:
        """Create a regular expression that matches the stroke with
        the acceptable annotations and given annotations according to
        the search options.

        :param stroke: The stroke for which to generate annotations.
        :param annotations: A collection of annotations used in generating
            the regular expression.

        :returns: A regular expression.
        """

        def pack(tup):
            if len(tup) == 1:
                return re.escape(tup[0]) + "?"
            return "[" + "".join(tup) + "]?"

        possible = grammar.ANNOTATIONS.get(stroke, list())

        if (
            self.annotation_mode is Strictness.MEDIUM
            or self.annotation_mode is Strictness.HIGH
        ):
            builder = []
            i = 0
            while i < len(possible):
                match = False
                for an in annotations:
                    if an in possible[i]:
                        builder.append(re.escape(an))
                        match = True
                        break
                if not match and self.annotation_mode is not Strictness.HIGH:
                    builder.append(pack(possible[i]))
                i += 1

            return stroke + "".join(builder)

        return stroke + "".join([pack(tup) for tup in possible])

    def make_uncertainty_regex(
        self, stroke: str, uncertainty: int, annotations: list = []
    ) -> str:
        """Create a regular expression that matches a stroke within a given
        uncertainty while applying provided annotations.

        :param stroke: The stroke for which to generate alternatives.
        :param uncertainty: The uncertainty to apply to the stroke.
        :param annotations: A list of annotations to use in the generation.
        :returns: A regular expression.
        """

        similars = get_similar(stroke, uncertainty)
        flattened = []
        for group in similars:
            for token in group:
                flattened.append(self.make_annotation_regex(token, annotations))

        return "(" + "|".join(flattened) + ")"

    def build_regex(self, interpretation: Interpretation) -> str:
        """Create a regular expression from a grascii interpretation based
        on the constructor search parameters.

        :param interpretation: The interpretation for which to generate a
            regular expression.
        :returns: A regular expression.
        """

        # aspirate and disjoiner are put in a regex group so they can be
        # compared later for calculating match metrics
        aspirate = "(" + re.escape(grammar.ASPIRATE) + ")"
        disjoiner = "(" + re.escape(grammar.DISJOINER) + ")"

        disjoiner_count = 0

        # grascii words are the first word of a dictionary entry which starts
        # at the beginning of the line
        builder = list("^")
        i = 0
        if self.search_mode is SearchMode.MATCH or self.search_mode is SearchMode.START:
            # start the matched_grascii group
            builder.append("(?P<matched_grascii>")

            # match up to two aspirates at the beginning of a word
            if self.aspirate_mode is Strictness.LOW:
                for j in range(2):
                    builder.append(aspirate)
                    builder.append("?")
            elif self.aspirate_mode is Strictness.MEDIUM:
                # retain existing aspirates and make others optional
                while i < len(interpretation) and interpretation[i] == grammar.ASPIRATE:
                    builder.append(aspirate)
                    i += 1
                if i < 2:
                    for j in range(2 - i):
                        builder.append(aspirate)
                        builder.append("?")
        elif self.search_mode is SearchMode.CONTAIN:
            # match any characters up to the end of the grascii word
            builder.append(r"\S*")

            # start the matched_grascii group
            builder.append("(?P<matched_grascii>")

        found_first = False

        while i < len(interpretation):
            token = interpretation[i]

            if token == grammar.ASPIRATE:
                if (
                    self.aspirate_mode is Strictness.MEDIUM
                    or self.aspirate_mode is Strictness.HIGH
                ):
                    # retain a non-optional aspirate
                    builder.append(aspirate)
                i += 1
                continue

            if token == grammar.DISJOINER:
                if (
                    self.disjoiner_mode is Strictness.MEDIUM
                    or self.disjoiner_mode is Strictness.HIGH
                ):
                    # retain a non-optional disjoiner
                    builder.append(disjoiner)
                disjoiner_count += 1
                i += 1
                continue

            uncertainty = self.uncertainty if found_first or not self.fix_first else 0
            if token in grammar.STROKES:
                last_character = builder[-1]
                if found_first:
                    if self.disjoiner_mode is Strictness.LOW or (
                        self.disjoiner_mode is Strictness.MEDIUM
                        and last_character != disjoiner
                    ):
                        # match optional disjoiner
                        builder.append(disjoiner)
                        builder.append("?")
                    if self.aspirate_mode is Strictness.LOW or (
                        self.aspirate_mode is Strictness.MEDIUM
                        and last_character != aspirate
                    ):
                        # match optional aspirate
                        builder.append(aspirate)
                        builder.append("?")
                found_first = True

                if token in grammar.ANNOTATIONS:
                    # this token may have annotations
                    if i + 1 < len(interpretation) and isinstance(
                        interpretation[i + 1], list
                    ):
                        builder.append(
                            self.make_uncertainty_regex(
                                token, uncertainty, interpretation[i + 1]
                            )
                        )
                        i += 2
                        continue

            builder.append(self.make_uncertainty_regex(token, uncertainty))
            i += 1

        if self.search_mode is SearchMode.MATCH:
            last_character = builder[-1]
            # match up to two aspirates at the end of the word
            if (
                self.aspirate_mode is Strictness.LOW
                or self.aspirate_mode is Strictness.MEDIUM
            ):
                if builder[-1] != aspirate:
                    for j in range(2):
                        builder.append(aspirate)
                        builder.append("?")
                elif builder[-2] != aspirate:
                    builder.append(aspirate)
                    builder.append("?")

            # match up to one disjoiner at the end of a word
            if self.disjoiner_mode is Strictness.LOW or (
                self.disjoiner_mode is Strictness.MEDIUM
                and disjoiner_count < 3
                and last_character != disjoiner
            ):
                builder.append(disjoiner)
                builder.append("?")

        # end the matched_grascii group
        builder.append(")")

        if self.search_mode is SearchMode.MATCH:
            # match the end of the word/line
            builder.append(r"(?:\Z|\s)")

        return "".join(builder)

    def get_starting_letters(self, interpretations: List[Interpretation]) -> Set[str]:
        """Get a set of starting letters based on the given interpretations
        factoring in uncertainty.

        :param interpretations: A list of interpretations to generate
            starting letters for.
        :returns: A set of characters.
        """

        if self.search_mode is SearchMode.CONTAIN:
            return grammar.HARD_CHARACTERS

        letters = set()
        for interp in interpretations:
            for token in interp:
                if token[0] in grammar.HARD_CHARACTERS:
                    if self.fix_first:
                        strokes = get_similar(token, 0)
                    else:
                        strokes = get_similar(token, self.uncertainty)
                    flattened_strokes = []
                    for tup in strokes:
                        flattened_strokes += [s for s in tup]
                    letters |= set(string[0] for string in flattened_strokes)
                    break

        return letters

    def generate_patterns_map(
        self, interpretations: List[Interpretation]
    ) -> List[Tuple[Interpretation, Pattern]]:
        """Generates a set of compiled regular expressions from a list
        of interpretations.

        :param interpretations: A list of interpretations to generate
            patterns for.
        :returns: A list of Interpretations with their corresponding
            Patterns.
        """

        patterns = []
        for interp in interpretations:
            regex = self.build_regex(interp)
            patterns.append((interp, re.compile(regex)))
        return patterns



"""A collection of useful information about the grascii grammar.

:var STROKES: A set of all valid strokes.
:var HARD_CHARACTERS: A set of all alphabetic characters that can appear as
    the first character in a stroke.
:var ANNOTATION_CHARACTERS: The set of all characters that are annotations.
:var ASPIRATE: The character corresponding to the aspirate.
:var MEDIUM_SOUND: The character corresponding to the medium sound of a standard
    vowel.
:var LONG_SOUND: The character corresponding to the long sound of a standard
    vowel.
:var LOOP: The character corresponding to the loop annotation.
:var REVERSE: The character corresponding to the reversing annotation.
:var WUNDERBAR: The character corresponding to the underbar (W) annotation.
:var ING: The character corresponding to the -ing ending.
:var LEFT: A character corresponding to a direction annotation.
:var RIGHT: A character corresponding to a direction annotation.
:var OBLIQUE: The character corresponding to the oblique annotation.
:var DISJOINER: The character corresponding to a disjoiner.
:var BOUNDARY: The character corresponding to a boundary.
:var INTERSECTION: The character corresponding to an intersection.
:var ALPHABET: The set of valid characters in the grascii language.
:var ANNOTATIONS: A dictionary of annotatable strokes to a sequence of
    acceptable annotations on the corresponding stroke. The sequece contains
    tuples of annotations. The tuples are ordered in the same order they
    must appear in a strict grascii string. The tuples contain mutually
    exclusive annotations. Ex: MEDIUM_SOUND and LONG_SOUND
"""

from __future__ import annotations

from typing import Dict, List, Tuple

_STROKES = "A AU A&' A&E B CH D DD DF DM DN DT DV E EU F G I J JND JNT K L LD \
    M MD MM MN MT N ND NG NK NT O OE P PND PNT R S SH SS T TD TH TM TN TV U V X XS Z"
_VOWELS = "A AU A&' A&E E EU I O OE U"

STROKES = set(_STROKES.split())
VOWELS = set(_VOWELS.split())
CONSONANTS = set(STROKES - VOWELS)

HARD_CHARACTERS = {c for c in "ABCDEFGIJKLMNOPRSTUVXZ"}
ANNOTATION_CHARACTERS = {c for c in ",.|~_()"}
ASPIRATE = "'"
MEDIUM_SOUND = "."
LONG_SOUND = ","
LOOP = "|"
REVERSE = "~"
WUNDERBAR = "_"
ING = "'"
LEFT = "("
RIGHT = ")"
OBLIQUE = ","
DISJOINER = "^"
BOUNDARY = "-"
INTERSECTION = "\\"

ALPHABET = set("ABCDEFGIJKLMNOPRSTUVXZ'.,|~_()^-&\\")

_CIRCLE_VOWEL_ANNOTATIONS: List[Tuple[str, ...]] = [
    (REVERSE,),
    (LOOP,),
    (MEDIUM_SOUND, LONG_SOUND),
    (WUNDERBAR,),
]
_CIRCLE_DIPHTHONG_ANNOTATIONS: List[Tuple[str, ...]] = [
    (REVERSE,),
    (LOOP,),
    (WUNDERBAR,),
]
_HOOK_DIPHTHONG_ANNOTATIONS: List[Tuple[str, ...]] = [(WUNDERBAR,)]
_DIRECTED_CONSONANT_ANNOTATIONS: List[Tuple[str, ...]] = [(LEFT, RIGHT), (OBLIQUE,)]

ANNOTATIONS: Dict[str, List[Tuple[str, ...]]] = {
    "A": _CIRCLE_VOWEL_ANNOTATIONS,
    "E": _CIRCLE_VOWEL_ANNOTATIONS,
    "O": [(LEFT,), (MEDIUM_SOUND, LONG_SOUND), (WUNDERBAR,)],
    "U": [(RIGHT,), (MEDIUM_SOUND, LONG_SOUND), (WUNDERBAR,)],
    "I": _CIRCLE_DIPHTHONG_ANNOTATIONS,
    "A&'": _CIRCLE_DIPHTHONG_ANNOTATIONS,
    "A&E": _CIRCLE_DIPHTHONG_ANNOTATIONS,
    "EU": _HOOK_DIPHTHONG_ANNOTATIONS,
    "AU": _HOOK_DIPHTHONG_ANNOTATIONS,
    "OE": _HOOK_DIPHTHONG_ANNOTATIONS,
    "S": _DIRECTED_CONSONANT_ANNOTATIONS,
    "Z": _DIRECTED_CONSONANT_ANNOTATIONS,
    "TH": _DIRECTED_CONSONANT_ANNOTATIONS,
    "SH": [(OBLIQUE,)],
}



from __future__ import annotations

APP_NAME = "grascii"
__version__ = "0.6.0"

from grascii.dictionary.build import (
    BuildMessage,
    BuildSummary,
    DictionaryBuilder,
    DictionaryOutputOptions,
)
from grascii.dictionary.pipeline import PipelineFunc
from grascii.parser import (
    GrasciiParser,
    GrasciiValidator,
    Interpretation,
    InvalidGrascii,
    interpretation_to_string,
)
from grascii.regen import SearchMode, Strictness
from grascii.searchers import GrasciiSearcher, RegexSearcher, ReverseSearcher, Searcher



"""
Contains metrics for comparing search queries to regular expression matches.
"""

from __future__ import annotations

import string
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
    Match,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    TypeVar,
)

from grascii import grammar
from grascii.parser import Interpretation
from grascii.similarities import get_similar

if TYPE_CHECKING:
    if sys.version_info >= (3, 8):
        from typing import Protocol
    else:
        from typing_extensions import Protocol
    from grascii.searchers import SearchResult

    class Comparable(Protocol):
        def __lt__(self, other: Any) -> bool:
            pass

    CT = TypeVar("CT", bound=Comparable)


class AnnotatedStroke(NamedTuple):
    stroke: str
    annotations: Set[str]


GrasciiSequence = List[AnnotatedStroke]


def interpretation_to_gsequence(interp: Interpretation) -> GrasciiSequence:
    """Convert an interpretation into a GrasciiSequence

    :param interp: The interpretation to convert.
    :returns: A GrasciiSequence.
    """

    sequence: GrasciiSequence = []
    i = 0
    while i < len(interp):
        item = interp[i]
        assert isinstance(item, str), item
        if i + 1 < len(interp) and isinstance(interp[i + 1], list):
            sequence.append(AnnotatedStroke(item, set(interp[i + 1])))
            i += 2
        else:
            sequence.append(AnnotatedStroke(item, set()))
            i += 1
    return sequence


def match_to_gsequence(match: Match[str]) -> GrasciiSequence:
    """Convert a match into a ``GrasciiSequence``

    :param match: The match to convert.
    :returns: A ``GrasciiSequence``.
    """

    sequence: GrasciiSequence = []
    adjusted_matched_grascii_index = match.re.groupindex["matched_grascii"] - 1
    for index, group in enumerate(match.groups()):
        if index == adjusted_matched_grascii_index:
            # the matched_grascii group is not a leaf group
            continue
        if group is None:
            continue
        i = 0
        while i < len(group) and group[i] in string.ascii_uppercase:
            i += 1
        if i > 0:
            sequence.append(AnnotatedStroke(group[:i], set(group[i:])))
        else:
            sequence.append(AnnotatedStroke(group, set()))
    return sequence


def gsequence_distance(seq1: GrasciiSequence, seq2: GrasciiSequence) -> int:
    """Compute a weighed Levenshtein distance between two sequences of annotated
    strokes.

    :param seq1: A ``GrasciiSequence``
    :param seq2: A second ``GrasciiSequence``
    :returns: A distance between seq1 and seq2.
    """

    BASE_COST = 4
    COSTS = {"~": 2, "|": 2, ".": 1, ",": 1, "(": 1, ")": 1, "_": 1}
    ASP_COST = 1
    DIS_COST = BASE_COST

    def compute_ins_del_cost(stroke: AnnotatedStroke) -> int:
        """Compute the insertion and deletion cost of a stroke and its
        annotations.

        :param stroke: A stroke and its annotations for which to calculate its cost.
        :returns: A cost of insertion/deletion.
        """

        cost = 0
        if stroke.stroke in grammar.STROKES:
            cost += BASE_COST
        elif stroke.stroke == grammar.ASPIRATE:
            cost += ASP_COST
        elif stroke.stroke == grammar.DISJOINER:
            cost += DIS_COST

        assert cost > 0, stroke
        cost += sum(COSTS.get(annotation, 0) for annotation in stroke.annotations)
        return cost

    def compute_sub_cost(stroke1: AnnotatedStroke, stroke2: AnnotatedStroke) -> int:
        """Compute the cost of substituting an annotated stroke with another one.

        :param stroke1: The stroke to replace.
        :param stroke2: The new stroke to add.
        :returns: A cost of substitution.
        """

        cost = 0
        if stroke1.stroke in grammar.STROKES and stroke2.stroke in grammar.STROKES:
            cost += get_distance(stroke1.stroke, stroke2.stroke) * BASE_COST
            diff = stroke1.annotations ^ stroke2.annotations
            cost += sum(COSTS.get(annotation, 0) for annotation in diff)
            return cost
        if stroke1.stroke == stroke2.stroke:
            return 0
        return max(compute_ins_del_cost(stroke1), compute_ins_del_cost(stroke2))

    def get_distance(stroke1: str, stroke2: str) -> int:
        """Calculate the distance between two strokes in a similarity graph.
        Restricted to 0-3.

        :param stroke1: A stroke
        :param stroke2: A second stroke to compute the distance between
        :returns: A distance.
        """

        def in_similarity_graph(distance: int) -> bool:
            similars = get_similar(stroke1, distance)
            similar_strokes = set()
            for tup in similars:
                similar_strokes |= set(tup)
            return stroke2 in similar_strokes

        for distance in range(3):
            if in_similarity_graph(distance):
                return distance
        return 3

    vec0 = [0]
    for stroke in seq2:
        vec0.append(vec0[-1] + compute_ins_del_cost(stroke))
    vec1 = [0 for i in range(len(seq2) + 1)]

    for stroke1 in seq1:
        vec1[0] = vec0[0] + compute_ins_del_cost(stroke1)

        for j, stroke2 in enumerate(seq2):
            del_cost = vec0[j + 1] + compute_ins_del_cost(stroke1)
            ins_cost = vec1[j] + compute_ins_del_cost(stroke2)
            sub_cost = vec0[j] + compute_sub_cost(stroke1, stroke2)
            vec1[j + 1] = min(del_cost, ins_cost, sub_cost)

        vec0, vec1 = vec1, vec0

    return vec0[len(seq2)]


IT = TypeVar("IT")


def determine_shortest_distance(
    matches: List[Tuple[IT, Match[str]]], func: Callable[[IT, Match[str]], CT]
) -> CT:
    """Determine the minimum value produced by a function on pairs of interpretations
    and matches. Utility for building metrics.
    """
    distance: Optional[CT] = None
    for interp, match in matches:
        new_distance = func(interp, match)
        if distance is None or new_distance < distance:
            distance = new_distance
    assert distance is not None
    return distance


def grascii_standard(result: SearchResult[Interpretation]) -> int:
    """Compute the standard metric for a grascii search.

    :param result: A ``SearchResult``
    :returns: A comparable key representing the distance between an ``Interpretation``
        and a ``Match``
    """

    def distance(interp: Interpretation, match: Match[str]) -> int:
        seq1 = interpretation_to_gsequence(interp)
        seq2 = match_to_gsequence(match)
        return gsequence_distance(seq1, seq2)

    min_start = min(r[1].start("matched_grascii") for r in result.matches)
    shortest_distance = determine_shortest_distance(result.matches, distance)

    # calculate an adjusted distance that will put "off-by-one" matches in the
    # same overall group as exact matches
    adjusted_distance = (
        shortest_distance if shortest_distance % 4 == 0 else shortest_distance - 1
    )

    return adjusted_distance, min_start, shortest_distance, len(result.entry.grascii)


def translation_standard(result: SearchResult[str]) -> Tuple[int, int]:
    """Compute the standard metric for a reverse search.

    :param result: A ``SearchResult``
    :returns: A comparable key representing the distance between a translation
        and a ``Match``
    """

    def distance(s: str, match: Match[str]) -> Tuple[int, int]:
        word_start = match.start("word") - match.end("grascii")
        return word_start, len(match.group("translation"))

    return determine_shortest_distance(result.matches, distance)


def trivial(result: SearchResult[IT]) -> int:
    """The trivial metric.

    :param result: A ``SearchResult``
    :returns: 0
    """
    return 0



"""
Acts as the main entry point for the grascii build command and contains the
dictionary builder implementation.

This can be invoked as a standalone program:
$ python -m grascii.dictionary.build --help
"""

from __future__ import annotations

import argparse
import logging
import os
import pathlib
import re
import sys
import time
from contextlib import nullcontext
from dataclasses import dataclass
from fileinput import FileInput
from typing import Dict, Iterable, List, NamedTuple, Optional, TextIO, Tuple

from grascii import grammar
from grascii.dictionary.pipeline import (
    CancelPipeline,
    PipelineFunc,
    create_grascii_check,
    create_spell_check,
    remove_boundaries,
    standardize_case,
)

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)


description = "Build a Grascii Dictionary"


def build_argparser(argparser: argparse.ArgumentParser) -> None:
    """Configure an ArgumentParser parser to parse the build command-line
    options.

    :param argparser: A fresh ArgumentParser to configure.
    """

    argparser.add_argument(
        "infiles", nargs="+", type=pathlib.Path, help="the files to process"
    )
    output_group = argparser.add_argument_group("output")
    exclusive_output_group = output_group.add_mutually_exclusive_group(required=True)
    exclusive_output_group.add_argument(
        "-o",
        "--output",
        type=os.path.abspath,
        help="path to a directory to output dictionary files",
    )
    exclusive_output_group.add_argument(
        "--no-output",
        action="store_true",
        help="do not output files and only perform validation",
    )
    output_group.add_argument(
        "-c",
        "--clean",
        action="store_true",
        help="clean the output directory before building",
    )
    validation_group = argparser.add_argument_group("validation")
    validation_group.add_argument(
        "-p",
        "--parse",
        action="store_true",
        help="enable syntax checking on grascii strings",
    )
    validation_group.add_argument(
        "-w",
        "--words",
        dest="words_file",
        type=pathlib.Path,
        help="path to a words file for spell checking",
    )
    validation_group.add_argument(
        "-n",
        "--count",
        dest="count_words",
        action="store_true",
        help="enable word count validation",
    )
    argparser.add_argument(
        "-v",
        "--verbose",
        dest="verbosity",
        action="count",
        default=0,
        help="increase output verbosity",
    )
    argparser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="suppress output",
    )


class NoMatchingOutputFile(Exception):
    pass


class BuildMessage(NamedTuple):
    file_name: str
    line: str
    line_number: int
    message: str
    level: int


class _BuilderLoggerAdapter(logging.LoggerAdapter):
    """A LoggerAdapter for dictionary builds that holds context on the current
    file and line being processed and maintains lists of BuildMessages
    for warnings and errors that occur during a build.
    """

    def __init__(self, logger: logging.Logger):
        super().__init__(logger, {})
        self._file_name: str = ""
        self._line: str = ""
        self._line_number: int = 0
        self.warnings: List[BuildMessage] = []
        self.errors: List[BuildMessage] = []

    def set_context(self, file_name: str, line: str, line_number: int):
        """Sets the context of the entry being processed.

        :param file_name: The name of the dictionary source file.
        :param line: The content of the line from the source file.
        :param line_number: The line number of ``line`` in the source file.
        """

        self._file_name = file_name
        self._line = line.strip()
        self._line_number = line_number

    def warning(self, msg, *args, **kwargs):
        message = f"{self._file_name}:{self._line_number} {msg}\n{self._line}"
        self.warnings.append(
            BuildMessage(
                self._file_name,
                self._line,
                self._line_number,
                message,
                logging.WARNING,
            )
        )
        self.logger.warning(message, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        message = f"{self._file_name}:{self._line_number} {msg}\n{self._line}"
        self.errors.append(
            BuildMessage(
                self._file_name,
                self._line,
                self._line_number,
                message,
                logging.ERROR,
            )
        )
        self.logger.error(message, *args, **kwargs)


@dataclass
class DictionaryOutputOptions:
    """Options for dictionary build output.

    :param output_dir: The directory where to output the dictionary files.
    :param clean: Whether to delete all files in the output directory before
        building.
    :param package: Whether to make the output directory a Python package by
        outputing an `__init__.py` file.
    :type output_dir: os.PathLike
    :type clean: bool
    :type package: bool
    """

    output_dir: os.PathLike
    clean: bool = False
    package: bool = False


class _OutputManager:
    def __init__(
        self,
        options: DictionaryOutputOptions,
        logger: logging.LoggerAdapter,
    ):
        self.options = options
        self._logger = logger
        self.entry_counts: Dict[str, int] = {}
        self._out_files: Dict[str, TextIO] = {}

    def __enter__(self):
        self._prepare_output_dir()
        return self._write_entry

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._close_output_files()

    def _prepare_output_dir(self) -> None:
        """Make and clean the output directory if necessary."""

        out_dir = pathlib.Path(self.options.output_dir)
        out_dir.mkdir(parents=True, exist_ok=True)
        if self.options.clean:
            self._logger.info("Cleaning output directory: %s", out_dir)
            for entry in out_dir.iterdir():
                entry.unlink()
        if self.options.package:
            self._logger.info("Creating __init__.py in %s", out_dir)
            out_dir.joinpath("__init__.py").touch()

    def _get_output_file(self, grascii: str) -> TextIO:
        """Get an output file corresponding to the first alphabetic characters
        in a grascii string.

        :param grascii: A grascii string to get an output file for.
        :returns: A text stream.
        """

        index = 0
        while index < len(grascii) and grascii[index] not in grammar.HARD_CHARACTERS:
            index += 1
        if index == len(grascii):
            raise NoMatchingOutputFile()

        char = grascii[index]
        try:
            result = self._out_files[char]
            self.entry_counts[char] += 1
            return result
        except KeyError:
            out_file = pathlib.Path(self.options.output_dir, char)
            self._out_files[char] = out_file.open("w")
            self._logger.info("Opened output file: %s", out_file)
            self.entry_counts[char] = 1
            return self._out_files[char]

    def _write_entry(self, grascii: str, translation: str) -> None:
        """Write an entry to an output file.

        :param grascii: The grascii string to write.
        :param translation: The grascii string's corresponding translation to write.
        """

        out = self._get_output_file(grascii)
        out.write(grascii + " ")
        out.write(translation + "\n")

    def _close_output_files(self) -> None:
        """Close all output files."""

        for f in self._out_files.values():
            f.close()
        self._logger.info("Closed output files")


@dataclass
class BuildSummary:
    """Results of a dictionary build.

    :param time: The duration of the build in seconds.
    :param warnings: A list of warnings that occurred during the build.
    :param errors: A list of errors that occurred during the build.
    :param output_dir: The output directory of the dictionary.
    :param entry_counts: A dictionary of file names in the output directory to
        the number of entries written to that file.
    :type time: float
    :type warnings: List[BuildMessage]
    :type errors: List[BuildMessage]
    :type output_dir: Optional[os.PathLike]
    :type entry_counts: Optional[Dict[str, int]]
    """

    time: float
    warnings: List[BuildMessage]
    errors: List[BuildMessage]
    output_dir: Optional[os.PathLike]
    entry_counts: Optional[Dict[str, int]]

    def __str__(self):
        builder = []

        total = 0
        if self.entry_counts is not None:
            for key, val in self.entry_counts.items():
                total += val
                builder.append(
                    f"Wrote {val} entries to {os.path.join(self.output_dir, key)}"
                )
        if total > 0:
            builder.append("")

        builder.append(f"Finished Build in {self.time:.5f} seconds")
        if self.entry_counts is not None:
            builder.append(f"Entries: {total}")
        builder.append(f"Warnings: {len(self.warnings)}")
        builder.append(f"Errors: {len(self.errors)}")

        return "\n".join(builder)


DEFAULT_PIPELINE: List[PipelineFunc] = [remove_boundaries, standardize_case]
"""The default pipeline used for dictionary builds."""


class DictionaryBuilder:

    """A class that runs the build process for a grascii dictionary.

    :param parse: Whether to enable parse checking of grascii strings.
    :param words_file: A Path to a words file for spell checking
    :param count_words: Whether to enable word count validation.
    :param verbosity: Increase the output verbosity
    :param quiet: Suppress output
    :type parse: bool
    :type words_file: pathlib.Path
    :type count_words: bool
    :type verbosity: int
    :type quiet: bool
    """

    def __init__(self, **kwargs) -> None:
        self.pipeline: List[PipelineFunc] = kwargs.get("pipeline", DEFAULT_PIPELINE)
        self.count_words: bool = kwargs.get("count_words", False)
        self._logger = _BuilderLoggerAdapter(logger)
        quiet: bool = kwargs.get("quiet", False)
        verbosity: int = kwargs.get("verbosity", 0)
        self._set_logging_level(quiet, verbosity)

    def _set_logging_level(self, quiet: bool, verbosity: int):
        if quiet:
            logger.setLevel(logging.CRITICAL)
        else:
            levels = [logging.WARNING, logging.INFO, logging.DEBUG]
            verbosity = min(verbosity, len(levels))
            if verbosity > 0:
                logger.setLevel(levels[verbosity])

    def _parse_line(self, line: str) -> Optional[Tuple[str, str]]:
        """Parse a dictionary source file line into a Grascii string and its
        translation.

        :returns: A Grascii string and a translation or None if the line is a
            comment or contains an error."""

        tokens = line.split()
        if not tokens:
            return None

        if tokens[0][0] == "#":
            return None

        if tokens[0] == "?":
            self._logger.warning("Uncertainty")
            tokens = tokens[1:]

        match = re.match(r"\*(\d+)", tokens[0])
        count = 1
        if match:
            tokens = tokens[1:]
            assert match[1], match
            count = int(match[1])
        if len(tokens) < count + 1:
            self._logger.error(
                f"Too few words: Expected: {count} Got: {len(tokens) - 1}",
            )
            return None
        if count and len(tokens) > count + 1:
            if self.count_words:
                self._logger.warning(
                    f"Too many words: Expected: {count} Got: {len(tokens) - 1}"
                )

        return tokens[0], " ".join(tokens[1:])

    def build(
        self, infiles: Iterable[os.PathLike], output: Optional[DictionaryOutputOptions]
    ) -> BuildSummary:
        """Run the build based on the build settings given in the constructor.

        :param infiles: A collection of file paths to dictionary source files.
        :param output: Options for dictionary output.
        :type infiles: Iterable[os.PathLike]
        :type output: Optional[DictionaryOutputOptions]
        :returns: A BuildSummary
        """

        # reset warnings and errors
        self._logger = _BuilderLoggerAdapter(logger)

        # time the build
        start_time = time.perf_counter()

        self._logger.info("Starting build")

        if not output:
            self._logger.info(
                "Only checking source files. Output files will not be generated."
            )

        out = (
            _OutputManager(output, self._logger)
            if output
            else nullcontext(lambda x, y: None)
        )

        with out as write_entry:
            with FileInput(infiles) as f:
                for line in f:
                    self._logger.set_context(f.filename(), line, f.filelineno())
                    parsed = self._parse_line(line)
                    if not parsed:
                        continue
                    grascii, translation = parsed

                    try:
                        for func in self.pipeline:
                            grascii, translation = func(
                                grascii, translation, self._logger
                            )
                    except CancelPipeline:
                        self._logger.info(
                            "Pipeline aborted for {grascii}, {translation}"
                        )
                        continue

                    try:
                        write_entry(grascii, translation)
                    except NoMatchingOutputFile:
                        self._logger.error(
                            f"No output file for {grascii} {translation}"
                        )
                        continue

        end_time = time.perf_counter()
        total_time = end_time - start_time
        self._logger.info("Build completed in %s seconds", total_time)

        return BuildSummary(
            time=total_time,
            warnings=self._logger.warnings,
            errors=self._logger.errors,
            output_dir=output.output_dir if output else None,
            entry_counts=out.entry_counts if isinstance(out, _OutputManager) else None,
        )


def cli_build(args: argparse.Namespace) -> None:
    """Run a build using arguments parsed from the command line.

    :param args: A namespace of parsed arguments.
    """

    pipeline: List[PipelineFunc] = []
    pipeline.extend(DEFAULT_PIPELINE)
    if args.words_file:
        pipeline.append(create_spell_check(args.words_file))
    if args.parse:
        pipeline.append(create_grascii_check())

    builder = DictionaryBuilder(
        pipeline=pipeline,
        count_words=args.count_words,
        verbosity=args.verbosity,
        quiet=args.quiet,
    )
    output_options = (
        DictionaryOutputOptions(args.output, args.clean) if args.output else None
    )
    summary = builder.build(args.infiles, output_options)

    if not args.quiet:
        if summary.warnings or summary.errors:
            print()
        print(summary)


def main() -> None:
    """Run a build using arguments retrieved from sys.argv."""

    argparser = argparse.ArgumentParser(description)
    build_argparser(argparser)
    args = argparser.parse_args(sys.argv[1:])
    cli_build(args)


if __name__ == "__main__":
    main()



"""
Contains the base class for Searchers as well as multiple concrete
implementations of it.
"""
from __future__ import annotations

import re
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    List,
    Match,
    NamedTuple,
    Optional,
    Pattern,
    Sequence,
    Set,
    Tuple,
    TypeVar,
)

from grascii import defaults, grammar, metrics, regen
from grascii.dictionary import Dictionary
from grascii.parser import GrasciiParser, Interpretation

if TYPE_CHECKING:
    from grascii.metrics import Comparable

IT = TypeVar("IT")


class DictionaryEntry(NamedTuple):
    grascii: str
    translation: str


class SearchResult(Generic[IT]):
    def __init__(
        self,
        matches: List[Tuple[IT, Match[str]]],
        entry: DictionaryEntry,
        dictionary: Dictionary,
    ) -> None:
        self.matches = matches
        self.entry = entry
        self.dictionary = dictionary


class Searcher(ABC, Generic[IT]):

    """An abstract base class for objects that search Grascii dictionaries.

    :param dictionaries: The dictionaries to search.
    :type dictionaries: List[str]
    """

    def __init__(self, **kwargs: Any) -> None:
        dictionaries = (
            kwargs.get("dictionaries")
            if kwargs.get("dictionaries")
            else defaults.SEARCH["Dictionary"].split()
        )
        self.dictionaries = [Dictionary.new(name) for name in dictionaries]

    def perform_search(
        self,
        patterns: Iterable[Tuple[IT, Pattern[str]]],
        starting_letters: Set[str],
    ) -> Iterable[SearchResult[IT]]:
        """Perform a search of a Grascii Dictionary.

        :param patterns: An iterable of interpretations and corresponding compiled
            regular expression patterns.
        :param starting_letters: A set of letters used to index the search in
            a Grascii Dictionary.
        :returns: An iterable of search results
        """
        for dictionary in self.dictionaries:
            for item in sorted(starting_letters):
                try:
                    dict_file = dictionary.open(item)
                except FileNotFoundError:
                    continue
                with dict_file:
                    for line in dict_file:
                        matches = []
                        for interp, pattern in patterns:
                            match = pattern.search(line)
                            if match:
                                matches.append((interp, match))
                        if matches:
                            grascii, translation = line.strip().split(maxsplit=1)
                            entry = DictionaryEntry(grascii, translation)
                            yield SearchResult(matches, entry, dictionary)

    @abstractmethod
    def search(self, **kwargs: Any) -> Optional[Iterable[SearchResult[IT]]]:
        """An abstract method that runs a search with the given search
        options and returns the results."""
        ...

    def sorted_search(
        self,
        metric: Callable[[SearchResult[IT]], Comparable] = metrics.trivial,
        **kwargs: Any,
    ) -> Sequence[SearchResult[IT]]:
        """Run a search with the given args and sort the search results by the
        given metric.
        """

        search_results = self.search(**kwargs)
        if search_results:
            return sorted(search_results, key=lambda r: metric(r))
        return []


class GrasciiSearcher(Searcher[Interpretation]):

    """A subclass of Searcher that performs a search given a Grascii string.

    :param dictionaries: The dictionaries to search.
    :type dictionaries: List[str]
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._parser = GrasciiParser()

    def extract_search_args(self, **kwargs: Any) -> None:
        """Get the relevant arguments for search."""

        self.uncertainty = kwargs.get(
            "uncertainty", defaults.SEARCH.getint("Uncertainty")
        )
        # handle enum conversion error?
        try:
            self.search_mode = regen.SearchMode(
                kwargs.get("search_mode", defaults.SEARCH["SearchMode"])
            )
        except ValueError:
            self.search_mode = regen.SearchMode(
                defaults.DEFAULTS["Search"]["SearchMode"]
            )
        try:
            self.annotation_mode = regen.Strictness(
                kwargs.get("annotation_mode", defaults.SEARCH["AnnotationMode"])
            )
        except ValueError:
            self.annotation_mode = regen.Strictness(
                defaults.DEFAULTS["Search"]["AnnotationMode"]
            )
        try:
            self.aspirate_mode = regen.Strictness(
                kwargs.get("aspirate_mode", defaults.SEARCH["AspirateMode"])
            )
        except ValueError:
            self.aspirate_mode = regen.Strictness(
                defaults.DEFAULTS["Search"]["AspirateMode"]
            )
        try:
            self.disjoiner_mode = regen.Strictness(
                kwargs.get("disjoiner_mode", defaults.SEARCH["DisjoinerMode"])
            )
        except ValueError:
            self.disjoiner_mode = regen.Strictness(
                defaults.DEFAULTS["Search"]["DisjoinerMode"]
            )
        self.fix_first = kwargs.get("fix_first", False)
        self.interpretation_mode = kwargs.get(
            "interpretation", defaults.SEARCH["Interpretation"]
        )

    def search(self, **kwargs: Any) -> Optional[Iterable[SearchResult[Interpretation]]]:
        """
        :param grascii: [Required] The grascii string to use in the search.
        :param uncertainty: The uncertainty of the grascii string.
        :param search_mode: The search mode to use.
        :param annotation_mode: How to handle annotations in the search.
        :param aspirate_mode: How to handle annotations in the search.
        :param disjoiner_mode: How to handle annotations in the search.
        :param fix_first: Apply an uncertainty of 0 to the first token.
        :param interpretation: How to handle ambiguous grascii strings.
        :type grascii: str
        :type uncertainty: int: 0, 1, or 2
        :type search_mode: str: one of regen.SearchMode values
        :type annotation_mode: str: one of regen.Strictness values
        :type aspirate_mode: str: one of regen.Strictness values
        :type disjoiner_mode: str: one of regen.Strictness values
        :type fix_first: bool
        :type interpretation: "best" or "all"
        :returns: A list of search results.
        :rtype: List[str]
        """

        grascii = kwargs["grascii"].upper()
        self.extract_search_args(**kwargs)

        interpretations = self._parser.interpret(grascii)

        builder = regen.RegexBuilder(
            uncertainty=self.uncertainty,
            search_mode=self.search_mode,
            aspirate_mode=self.aspirate_mode,
            annotation_mode=self.annotation_mode,
            disjoiner_mode=self.disjoiner_mode,
            fix_first=self.fix_first,
        )

        interps = (
            [next(interpretations)]
            if self.interpretation_mode == "best"
            else list(interpretations)
        )
        patterns = builder.generate_patterns_map(interps)
        starting_letters = builder.get_starting_letters(interps)

        return self.perform_search(patterns, starting_letters)

    def sorted_search(
        self,
        metric: Callable[
            [SearchResult[Interpretation]], Comparable
        ] = metrics.grascii_standard,
        **kwargs: Any,
    ) -> Sequence[SearchResult[Interpretation]]:
        return super().sorted_search(metric, **kwargs)


class RegexSearcher(Searcher[str]):

    """A subclass of Searcher that searches a grascii dictionary given
    a raw regular expression.

    :param dictionaries: The dictionaries to search.
    :type dictionaries: List[str]
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def search(self, **kwargs: Any) -> Iterable[SearchResult[str]]:
        """
        :param regexp: [Required] A regular expression to use in a search.
        :returns: A list of search results.
        :rtype: List[str]
        """

        regex = kwargs["regexp"]
        pattern = re.compile(regex)
        patterns = [(pattern.pattern, pattern)]

        starting_letters = grammar.HARD_CHARACTERS
        return self.perform_search(patterns, starting_letters)


class ReverseSearcher(RegexSearcher):

    """A subclass of RegexSearcher that searches a grascii dictionary
    given a word.

    :param dictionaries: The dictionaries to search.
    :type dictionaries: List[str]
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def search(self, **kwargs: Any) -> Iterable[SearchResult[str]]:
        """
        :param reverse: [Required] A word to search for.
        :returns: A list of search results.
        :rtype: List[str]
        """

        word = kwargs["reverse"]
        escaped_word = re.escape(word)
        kwargs["regexp"] = (
            r"(?i)(?P<grascii>.+?\s+)"
            + f"(?P<translation>(.*\\s)?(?P<word>{escaped_word}).*)(\\s|\\Z)"
        )

        return super().search(**kwargs)

    def sorted_search(
        self,
        metric: Callable[
            [SearchResult[str]], Comparable
        ] = metrics.translation_standard,
        **kwargs: Any,
    ) -> Sequence[SearchResult[str]]:
        return super().sorted_search(metric, **kwargs)



from __future__ import annotations

import argparse
import sys

from grascii.dictionary import build_argparser


def main() -> None:
    argparser = argparse.ArgumentParser(prog="dictionary")
    build_argparser(argparser)
    args = argparser.parse_args(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()



from __future__ import annotations

import unittest

from grascii.outline import Outline
from grascii.parser import GrasciiParser, interpretation_to_string


class TestInferredDirections(unittest.TestCase):
    parser = GrasciiParser()

    def run_tests(self, tests):
        for test in tests:
            with self.subTest(word=test[0], expected=test[1]):
                interpretation = next(
                    self.parser.interpret(test[0], preserve_boundaries=True)
                )
                outline = Outline(interpretation)
                self.assertEqual(str(outline), test[1])

    def test_no_directed_consonants(self):
        tests = [
            ("DREM", "DREM"),
            ("JENERK", "JENERK"),
            ("GAT", "GAT"),
            ("LASH", "LASH"),
            ("TRELN", "TRELN"),
        ]
        self.run_tests(tests)

    def test_lone_s(self):
        tests = [("S", "S)"), ("S)", "S)"), ("S(", "S(")]
        self.run_tests(tests)

    def test_s_joined_to_curves(self):
        tests = [
            ("SPRA", "S(PRA"),
            ("REPS", "REPS("),
            ("PAS", "PAS("),
            ("SFER", "S)FER"),
            ("SAF", "S)AF"),
            ("FAS", "FAS)"),
            ("SKAT", "S)KAT"),
            ("SEK", "S)EK"),
            ("MAKS", "MAKS)"),
            ("KAS", "KAS)"),
            ("SLA", "S(LA"),
            ("SALS", "S(ALS("),
        ]
        self.run_tests(tests)

    def test_s_joined_to_forward_lines(self):
        tests = [
            ("STA", "S)TA"),
            ("SET", "S)ET"),
            ("NETS", "NETS("),
            ("SED", "S)ED"),
            ("ODS", "ODS("),
            ("DAS", "DAS("),
            ("SNO", "S)NO"),
            ("SEN", "S)EN"),
            ("SMAK", "S)MAK"),
            ("SAM", "S)AM"),
            ("LENS", "LENS("),
            ("NES", "NES("),
        ]
        self.run_tests(tests)

    def test_s_joined_to_downward_lines(self):
        tests = [("SASH", "S)ASH"), ("SAJ", "S)AJ"), ("CHES", "CHES)")]
        self.run_tests(tests)

    def test_s_o(self):
        tests = [
            ("SO", "S)O"),
            ("SORO", "S)ORO"),
            ("SOL", "S)OL"),
            ("SOFA", "S)OFA"),
            ("SOP", "S)OP"),
            ("SOD", "S)OD"),
        ]
        self.run_tests(tests)

    def test_s_u(self):
        tests = [
            ("US", "US)"),
            ("BUS", "BUS)"),
            ("FUS", "FUS)"),
            ("GUST", "GUS)T"),
            ("GRASHUS", "GRASHUS)"),
            ("VESHUS", "VESHUS)"),
        ]
        self.run_tests(tests)

    def test_sandwiched_s(self):
        tests = [
            ("UOSP", "UOS(P"),
            ("PAST", "PAS(T"),
            ("UESP", "UES(P"),
            ("UESTF", "UES)TF"),
            ("GOSEP", "GOS(EP"),
            ("TRESPAS", "TRES(PAS("),
            ("TRESL", "TRES(L"),
            ("VEST", "VES)T"),
            ("KASK", "KAS)K"),
            ("KASM", "KAS)M"),
            ("RESK", "RES(K"),
            ("ASAL", "AS(AL"),
            ("DOSEL", "DOS(EL"),
            ("FLASK", "FLAS(K"),
            ("KLASP", "KLAS(P"),
        ]
        self.run_tests(tests)

    def test_sandwiched_s2(self):
        tests = [
            ("VESTRE", "VES)TRE"),
            ("OFSET", "OFS)ET"),
            ("BOST", "BOS(T"),
            ("TAST", "TAS(T"),
            ("DESK", "DES(K"),
            ("MASK", "MAS(K"),
        ]
        self.run_tests(tests)

    def test_s_after_i(self):
        tests = [
            ("ISI", "IS(I"),
            ("I-SI", "I-S)I"),
            ("ISBERG", "IS(BERG"),
            ("ISKREM", "IS(KREM"),
            ("IS^K", "IS(^K"),
            ("ISOLA", "IS(OLA"),
            ("I-SOLASH", "I-S)OLASH"),
            ("NIS", "NIS("),
            ("VIS", "VIS)"),
            ("UIS", "UIS)"),
            ("U-IS", "U-IS("),
        ]
        self.run_tests(tests)

    def test_z(self):
        tests = [
            ("GAZ,", "GAZ),"),
            ("FAZ,", "FAZ),"),
            ("ZELUS", "Z(ELUS)"),
            ("ZERO", "Z(ERO"),
            ("MAZ", "MAZ("),
            ("RAZ", "RAZ("),
            ("DOZ", "DOZ("),
            ("LAZE", "LAZ(E"),
        ]
        self.run_tests(tests)

    def test_lone_th(self):
        tests = [("TH", "TH("), ("TH)", "TH)"), ("TH(", "TH(")]
        self.run_tests(tests)

    def test_th_joinings(self):
        tests = [
            ("THEK", "TH(EK"),
            ("THEM", "TH(EM"),
            ("DUTH", "DUTH("),
            ("THO", "TH)O"),
            ("THRO", "TH)RO"),
            ("ATHLET", "ATH)LET"),
            ("MOTH", "MOTH)"),
            ("ERTH", "ERTH)"),
            ("'ELTH", "'ELTH)"),
        ]
        self.run_tests(tests)

    def test_s_th_words(self):
        tests = [
            ("AS", "AS)"),
            ("SE", "S)E"),
            ("ESA", "ES)A"),
            ("'ETH", "'ETH("),
            ("'ATH", "'ATH("),
            ("THE", "TH(E"),
            ("THES", "TH(ES)"),
            ("SES", "S)ES)"),
            ("SETH", "S)ETH("),
        ]
        self.run_tests(tests)

    def test_lone_o(self):
        tests = [("O", "O"), ("O(", "O(")]
        self.run_tests(tests)

    def test_o_joinings(self):
        tests = [
            ("ON", "O(N"),
            ("OR", "O(R"),
            ("MON", "MO(N"),
            ("'OL", "'O(L"),
            ("DOM", "DO(M"),
            ("NOM", "NO(M"),
            ("MOD", "MOD"),
            ("SHOL", "SHOL"),
            ("FOM", "FOM"),
            ("BOCH", "BOCH"),
            ("ROG", "ROG"),
            ("PO,L", "PO,L"),
        ]
        self.run_tests(tests)

    def test_lone_u(self):
        tests = [
            ("U", "U"),
            ("U)", "U)"),
        ]
        self.run_tests(tests)

    def test_u_joinings(self):
        tests = [
            ("NUN", "NU)N"),
            ("MUD", "MU)D"),
            ("MUF", "MU)F"),
            ("MUN", "MU)N"),
            ("MUG", "MU)G"),
            ("MUD", "MU)D"),
            ("KUL", "KU)L"),
            ("GUL", "GU)L"),
            ("FUT", "FUT"),
            ("'UK", "'UK"),
            ("PUF", "PUF"),
            ("JUG", "JUG"),
            ("TUF", "TUF"),
            ("KUF", "KUF"),
            ("GUSH", "GUSH"),
        ]
        self.run_tests(tests)

    def test_s_boundaries(self):
        tests = [
            ("LSN", "LS(N"),
            ("L-SN", "L-S)N"),
            ("LOBSTE~", "LOBS(TE~"),
            ("LOB-STE~", "LOB-S)TE~"),
            ("LONSM", "LO(NS(M"),
            ("LON-SM", "LO(N-S)M"),
            ("LESR", "LES(R"),
            ("LESN", "LES(N"),
            ("LE-SN", "LE-S)N"),
            ("RSN", "RS(N"),
            ("R-SN", "R-S)N"),
            ("RSED", "RS(ED"),
            ("R-SED", "R-S)ED"),
            ("DSET", "DS(ET"),
            ("D-SET", "D-S)ET"),
            ("OS", "OS("),
            ("O-S", "O-S)"),
        ]
        self.run_tests(tests)

    def test_o_boundaries(self):
        tests = [
            ("NOR", "NO(R"),
            ("NO-R", "NO-R"),
            ("TOL", "TO(L"),
            ("TO-L", "TO-L"),
            ("CHOR", "CHOR"),
            ("CHO-R", "CHO-R"),
            ("CH-OR", "CH-O(R"),
            ("GON", "GO(N"),
            ("GO-N", "GO-N"),
            ("JON", "JON"),
            ("JO-N", "JO-N"),
            ("J-ON", "J-O(N"),
            ("DOM", "DO(M"),
            ("DO-M", "DO-M"),
        ]
        self.run_tests(tests)

    def test_u_boundaries(self):
        tests = [
            ("NUN", "NU)N"),
            ("N-UN", "N-UN"),
            ("MUT", "MU)T"),
            ("M-UT", "M-UT"),
            ("MNU", "MNU)"),
            ("MN-U", "MN-U"),
            ("KUT", "KUT"),
            ("KUR", "KU)R"),
            ("KU-R", "KU-R"),
            ("GUL", "GU)L"),
            ("G-UL", "G-UL"),
            ("GU-L", "GU-L"),
        ]
        self.run_tests(tests)

    def test_th_boundaries(self):
        tests = [
            ("THR", "TH)R"),
            ("TH-R", "TH(-R"),
            ("OTH", "OTH)"),
            ("O-TH", "O-TH("),
            ("THL", "TH)L"),
            ("TH-L", "TH(-L"),
            ("RTHO", "RTH)O"),
            ("R-THO", "R-TH)O"),
            ("RTH-O", "RTH)-O"),
            ("R-TH-O", "R-TH(-O"),
        ]
        self.run_tests(tests)

    def test_th_circle_vowel(self):
        tests = [
            ("THER", "TH)ER"),
            ("THE-R", "TH(E-R"),
            ("THIROE", "TH)IROE"),
            ("TH-IROE", "TH(-IROE"),
            ("LATH", "LATH)"),
            ("LA-TH", "LA-TH("),
            ("RATH", "RATH)"),
            ("R-ATH", "R-ATH("),
            ("THA&ER", "TH)A&ER"),
            ("LA&'TH", "LA&'TH)"),
        ]
        self.run_tests(tests)


class TestToInterpretation(unittest.TestCase):
    parser = GrasciiParser()

    def test_to_interpretation(self):
        tests = [
            ("JENERK", "J-E-N-E-R-K"),
            ("GAT", "G-A-T"),
            ("LASH", "L-A-SH"),
            ("MON", "M-O(-N"),
            ("'OL", "'-O(-L"),
            ("L-SN", "L-S)-N"),
            ("LOBSTE~", "L-O-B-S(-T-E~"),
            ("ESA", "E-S)-A"),
            ("'ETH", "'-E-TH("),
            ("TRESPAS", "T-R-E-S(-P-A-S("),
        ]
        for test in tests:
            with self.subTest(word=test[0], expected=test[1]):
                interpretation = next(
                    self.parser.interpret(test[0], preserve_boundaries=True)
                )
                outline = Outline(interpretation)
                self.assertEqual(
                    interpretation_to_string(outline.to_interpretation()), test[1]
                )



from __future__ import annotations

import re
import unittest
from typing import List, Tuple

from grascii import grammar, regen


class TestAnnotationRegex(unittest.TestCase):
    def check_strictness_low(
        self, annotations: List[str], texts: List[Tuple[str, str]]
    ):
        builder = regen.RegexBuilder(annotation_mode=regen.Strictness.LOW)
        for a in annotations:
            for t in texts:
                regex = builder.make_annotation_regex(t[0], a)
                self.assertRegex(t[0] + t[1], regex)

    def test_strictness_low_circle_vowel(self):
        annotations = [[], [","], ["~", ","]]
        texts = [
            ("A", ""),
            ("A", "~"),
            ("A", "|."),
            ("A", "~|,_"),
            ("E", ""),
            ("E", "~"),
            ("E", "|."),
            ("E", "~|,_"),
        ]
        self.check_strictness_low(annotations, texts)

    def test_strictness_low_hook_vowel(self):
        annotations_o = [[], [","], ["(", ","], ["(", "_"]]
        texts_o = [("O", ""), ("O", "("), ("O", "(,"), ("O", "(._")]
        annotations_u = [[], [","], [")", ","], [")", "_"]]
        texts_u = [("U", ""), ("U", ")"), ("U", "),"), ("U", ")._")]
        self.check_strictness_low(annotations_o, texts_o)
        self.check_strictness_low(annotations_u, texts_u)

    def test_strictness_low_circle_diphthong(self):
        annotations = [
            [],
            ["~"],
            ["~", "|"],
            ["_"],
        ]
        texts = [
            ("I", ""),
            ("I", "~"),
            ("I", "_"),
            ("I~|_"),
            ("A&E", ""),
            ("A&E", "~"),
            ("A&E", "_"),
            ("A&E~|_"),
            ("A'E", ""),
            ("A'E", "~"),
            ("A'E", "_"),
            ("A'E~|_"),
        ]
        self.check_strictness_low(annotations, texts)

    def test_strictness_low_hook_diphthong(self):
        annotations = [
            [],
            ["_"],
        ]
        texts = [
            ("AU", ""),
            ("AU", "_"),
            ("OE", ""),
            ("OE", "_"),
            ("EU", ""),
            ("EU", "_"),
        ]
        self.check_strictness_low(annotations, texts)

    def test_strictness_low_directed_consonant(self):
        annotations = [[], ["("], [")"], [","], ["(,"]]
        texts = [
            ("S", ""),
            ("S", "("),
            ("S", "),"),
            ("Z", ""),
            ("Z", "("),
            ("Z", "),"),
            ("TH", ""),
            ("TH", "("),
            ("TH", "),"),
        ]
        self.check_strictness_low(annotations, texts)

    def test_strictness_low_oblique_consonant(self):
        annotations = [[], [","]]
        texts = [("SH", ""), ("SH", ",")]
        self.check_strictness_low(annotations, texts)

    def check_strictness_medium(self, strokes, tests):
        builder = regen.RegexBuilder(annotation_mode=regen.Strictness.MEDIUM)
        for stroke in strokes:
            for test in tests:
                annotations = test[0]
                for text, expected in test[1]:
                    regex = builder.make_annotation_regex(stroke, annotations)
                    with self.subTest(annotations=annotations, text=stroke + text):
                        if expected:
                            self.assertRegex(stroke + text, regex)
                        else:
                            self.assertNotRegex(stroke + text, regex)

    def test_strictness_medium_circle_vowel(self):
        circle_vowels = ["A", "E"]
        tests = [
            ([], [("", True), (",", True), ("~|._", True)]),
            ([","], [("", False), (".", False), (",", True), ("~|,_", True)]),
            (["~"], [("", False), ("~", True), (".", False), ("~|._", True)]),
            (
                ["~", "|", ".", "_"],
                [
                    ("~", False),
                    ("|", False),
                    (".", False),
                    ("_", False),
                    ("~|._", True),
                ],
            ),
        ]
        self.check_strictness_medium(circle_vowels, tests)

    def test_strictness_medium_hook_vowel(self):
        tests1 = [
            ([], [("", True), (",", True), ("(,_", True)]),
            ([","], [("", False), (".", False), (",", True), ("(,_", True)]),
            (["("], [("", False), (".", False), ("(,_", True)]),
            (["(", ".", "_"], [(".", False), ("_", False), ("(._", True)]),
        ]
        tests2 = [
            ([], [("", True), (",", True), ("),_", True)]),
            ([","], [("", False), (".", False), (",", True), ("),_", True)]),
            ([")"], [("", False), (".", False), ("),_", True)]),
            ([")", ".", "_"], [(".", False), ("_", False), (")._", True)]),
        ]
        self.check_strictness_medium(["O"], tests1)
        self.check_strictness_medium(["U"], tests2)

    def test_strictness_medium_circle_diphthong(self):
        strokes = ["I", "A&E", "A&'"]
        tests = [
            ([], [("", True), ("~", True), ("|", True), ("_", True), ("~|_", True)]),
            (
                ["~"],
                [("", False), ("~", True), ("|", False), ("_", False), ("~|_", True)],
            ),
            (
                ["|"],
                [("", False), ("~", False), ("|", True), ("_", False), ("~|_", True)],
            ),
            (
                ["_"],
                [("", False), ("~", False), ("|", False), ("_", True), ("~|_", True)],
            ),
            (
                ["~", "|", "_"],
                [("", False), ("~", False), ("|", False), ("_", False), ("~|_", True)],
            ),
        ]
        self.check_strictness_medium(strokes, tests)

    def test_strictness_medium_hook_diphthong(self):
        strokes = ["AU", "OE", "EU"]
        tests = [
            ([], [("", True), ("_", True)]),
            (["_"], [("", False), ("_", True)]),
        ]
        self.check_strictness_medium(strokes, tests)

    def test_strictness_medium_directed_consonant(self):
        strokes = ["S", "Z", "TH"]
        tests = [
            (
                [],
                [
                    ("", True),
                    ("(", True),
                    (")", True),
                    (",", True),
                    ("(,", True),
                    ("),", True),
                ],
            ),
            (
                ["("],
                [
                    ("", False),
                    ("(", True),
                    (")", False),
                    (",", False),
                    ("(,", True),
                    ("),", False),
                ],
            ),
            (
                [")"],
                [
                    ("", False),
                    ("(", False),
                    (")", True),
                    (",", False),
                    ("(,", False),
                    ("),", True),
                ],
            ),
            (
                ["(", ","],
                [
                    ("", False),
                    ("(", False),
                    (")", False),
                    (",", False),
                    ("(,", True),
                    ("),", False),
                ],
            ),
            (
                [")", ","],
                [
                    ("", False),
                    ("(", False),
                    (")", False),
                    (",", False),
                    ("(,", False),
                    ("),", True),
                ],
            ),
        ]
        self.check_strictness_medium(strokes, tests)

    def test_strictness_medium_oblique_consonant(self):
        tests = [
            ([], [("", True), (",", True)]),
            ([","], [("", False), (",", True)]),
        ]
        self.check_strictness_medium(["SH"], tests)

    def check_strictness_high(self, strokes, tests):
        builder = regen.RegexBuilder(annotation_mode=regen.Strictness.HIGH)
        for stroke in strokes:
            for test in tests:
                annotations = test[0]
                for text, expected in test[1]:
                    regex = builder.make_annotation_regex(stroke, annotations)
                    with self.subTest(annotations=annotations, text=stroke + text):
                        if expected:
                            self.assertTrue(re.fullmatch(regex, stroke + text))
                        else:
                            self.assertFalse(re.fullmatch(regex, stroke + text))

    def test_strictness_high_circle_vowel(self):
        circle_vowels = ["A", "E"]
        tests = [
            ([], [("", True), (",", False), ("~|._", False)]),
            ([","], [("", False), (".", False), (",", True), ("~|,_", False)]),
            (["~"], [("", False), ("~", True), (".", False), ("~|._", False)]),
            (
                ["~", "|", ".", "_"],
                [
                    ("~", False),
                    ("|", False),
                    (".", False),
                    ("_", False),
                    ("~|._", True),
                ],
            ),
        ]
        self.check_strictness_high(circle_vowels, tests)

    def test_strictness_high_hook_vowel(self):
        tests1 = [
            ([], [("", True), (",", False), ("(,_", False)]),
            ([","], [("", False), (".", False), (",", True), ("(,_", False)]),
            (["("], [("", False), (".", False), ("(,_", False)]),
            (["(", ".", "_"], [(".", False), ("_", False), ("(._", True)]),
        ]
        tests2 = [
            ([], [("", True), (",", False), ("),_", False)]),
            ([","], [("", False), (".", False), (",", True), ("),_", False)]),
            ([")"], [("", False), (".", False), ("),_", False)]),
            ([")", ".", "_"], [(".", False), ("_", False), (")._", True)]),
        ]
        self.check_strictness_high(["O"], tests1)
        self.check_strictness_high(["U"], tests2)

    def test_strictness_high_circle_diphthong(self):
        strokes = ["I", "A&E", "A&'"]
        tests = [
            (
                [],
                [("", True), ("~", False), ("|", False), ("_", False), ("~|_", False)],
            ),
            (
                ["~"],
                [("", False), ("~", True), ("|", False), ("_", False), ("~|_", False)],
            ),
            (
                ["|"],
                [("", False), ("~", False), ("|", True), ("_", False), ("~|_", False)],
            ),
            (
                ["_"],
                [("", False), ("~", False), ("|", False), ("_", True), ("~|_", False)],
            ),
            (
                ["~", "|", "_"],
                [("", False), ("~", False), ("|", False), ("_", False), ("~|_", True)],
            ),
        ]
        self.check_strictness_high(strokes, tests)

    def test_strictness_high_hook_diphthong(self):
        strokes = ["AU", "OE", "EU"]
        tests = [
            ([], [("", True), ("_", False)]),
            (["_"], [("", False), ("_", True)]),
        ]
        self.check_strictness_high(strokes, tests)

    def test_strictness_high_directed_consonant(self):
        strokes = ["S", "Z", "TH"]
        tests = [
            (
                [],
                [
                    ("", True),
                    ("(", False),
                    (")", False),
                    (",", False),
                    ("(,", False),
                    ("),", False),
                ],
            ),
            (
                ["("],
                [
                    ("", False),
                    ("(", True),
                    (")", False),
                    (",", False),
                    ("(,", False),
                    ("),", False),
                ],
            ),
            (
                [")"],
                [
                    ("", False),
                    ("(", False),
                    (")", True),
                    (",", False),
                    ("(,", False),
                    ("),", False),
                ],
            ),
            (
                ["(", ","],
                [
                    ("", False),
                    ("(", False),
                    (")", False),
                    (",", False),
                    ("(,", True),
                    ("),", False),
                ],
            ),
            (
                [")", ","],
                [
                    ("", False),
                    ("(", False),
                    (")", False),
                    (",", False),
                    ("(,", False),
                    ("),", True),
                ],
            ),
        ]
        self.check_strictness_high(strokes, tests)

    def test_strictness_high_oblique_consonant(self):
        tests = [
            ([], [("", True), (",", False)]),
            ([","], [("", False), (",", True)]),
        ]
        self.check_strictness_high(["SH"], tests)

    def test_out_of_order_annotations(self):
        pass

    def test_invalid_annotations(self):
        pass

    def test_mutually_exclusive_annotations(self):
        pass


class RegexBuilderTester(unittest.TestCase):
    def run_tests(self, builder, tests):
        for test in tests:
            interp = test[0]
            for text, expected in test[1]:
                regex = builder.build_regex(interp)
                with self.subTest(interpretation=interp, text=text):
                    if expected:
                        self.assertRegex(text, regex)
                    else:
                        self.assertNotRegex(text, regex)


class TestDisjoiners(RegexBuilderTester):
    def test_strictness_low(self):
        builder = regen.RegexBuilder(disjoiner_mode=regen.Strictness.LOW)
        tests = [
            (["A", "B"], [("AB", True), ("A^B", True)]),
            (["A", "^", "B"], [("AB", True), ("A^B", True), ("A^^B", False)]),
            (
                ["A", "B", "D"],
                [
                    ("ABD", True),
                    ("A^BD", True),
                    ("AB^D", True),
                    ("A^B^D", True),
                    ("A^B^D^", True),
                ],
            ),
            (
                ["A", "^", "B", "D"],
                [
                    ("ABD", True),
                    ("A^BD", True),
                    ("AB^D", True),
                    ("A^B^D", True),
                    ("A^B^D^", True),
                ],
            ),
            (
                ["A", "B", "^", "D"],
                [
                    ("ABD", True),
                    ("A^BD", True),
                    ("AB^D", True),
                    ("A^B^D", True),
                    ("A^B^D^", True),
                ],
            ),
            (
                ["A", "^", "B", "^", "D"],
                [
                    ("ABD", True),
                    ("A^BD", True),
                    ("AB^D", True),
                    ("A^B^D", True),
                    ("A^B^D^", True),
                ],
            ),
            (
                ["A", "^", "B", "^", "D", "^"],
                [
                    ("ABD", True),
                    ("A^BD", True),
                    ("AB^D", True),
                    ("A^B^D", True),
                    ("A^B^D^", True),
                ],
            ),
        ]
        self.run_tests(builder, tests)

    def test_strictness_medium(self):
        builder = regen.RegexBuilder(disjoiner_mode=regen.Strictness.MEDIUM)
        tests = [
            (["A", "B"], [("AB", True), ("A^B", True)]),
            (["A", "^", "B"], [("AB", False), ("A^B", True), ("A^^B", False)]),
            (
                ["A", "B", "D"],
                [
                    ("ABD", True),
                    ("A^BD", True),
                    ("AB^D", True),
                    ("A^B^D", True),
                    ("A^B^D^", True),
                ],
            ),
            (
                ["A", "^", "B", "D"],
                [
                    ("ABD", False),
                    ("A^BD", True),
                    ("AB^D", False),
                    ("A^B^D", True),
                    ("A^B^D^", True),
                ],
            ),
            (
                ["A", "B", "^", "D"],
                [
                    ("ABD", False),
                    ("A^BD", False),
                    ("AB^D", True),
                    ("A^B^D", True),
                    ("A^B^D^", True),
                ],
            ),
            (
                ["A", "^", "B", "^", "D"],
                [
                    ("ABD", False),
                    ("A^BD", False),
                    ("AB^D", False),
                    ("A^B^D", True),
                    ("A^B^D^", True),
                ],
            ),
            (
                ["A", "^", "B", "^", "D", "^"],
                [
                    ("ABD", False),
                    ("A^BD", False),
                    ("AB^D", False),
                    ("A^B^D", False),
                    ("A^B^D^", True),
                ],
            ),
        ]
        self.run_tests(builder, tests)

    def test_strictness_high(self):
        builder = regen.RegexBuilder(disjoiner_mode=regen.Strictness.HIGH)
        tests = [
            (["A", "B"], [("AB", True), ("A^B", False)]),
            (["A", "^", "B"], [("AB", False), ("A^B", True), ("A^^B", False)]),
            (
                ["A", "B", "D"],
                [("ABD", True), ("A^BD", False), ("AB^D", False), ("A^B^D", False)],
                ("A^B^D^", False),
            ),
            (
                ["A", "^", "B", "D"],
                [
                    ("ABD", False),
                    ("A^BD", True),
                    ("AB^D", False),
                    ("A^B^D", False),
                    ("A^B^D^", False),
                ],
            ),
            (
                ["A", "B", "^", "D"],
                [
                    ("ABD", False),
                    ("A^BD", False),
                    ("AB^D", True),
                    ("A^B^D", False),
                    ("A^B^D^", False),
                ],
            ),
            (
                ["A", "^", "B", "^", "D"],
                [
                    ("ABD", False),
                    ("A^BD", False),
                    ("AB^D", False),
                    ("A^B^D", True),
                    ("A^B^D^", False),
                ],
            ),
            (
                ["A", "^", "B", "^", "D", "^"],
                [
                    ("ABD", False),
                    ("A^BD", False),
                    ("AB^D", False),
                    ("A^B^D", False),
                    ("A^B^D^", True),
                ],
            ),
        ]
        self.run_tests(builder, tests)

    def test_prefixes_strictness_low(self):
        builder = regen.RegexBuilder(disjoiner_mode=regen.Strictness.LOW)
        tests = [
            (["U"], [("U^", True), ("U", True)]),
            (["U", "^"], [("U^", True), ("U", True), ("U^^", False)]),
        ]
        self.run_tests(builder, tests)

    def test_prefixes_strictness_medium(self):
        builder = regen.RegexBuilder(disjoiner_mode=regen.Strictness.MEDIUM)
        tests = [
            (["U"], [("U^", True), ("U", True)]),
            (["U", "^"], [("U^", True), ("U", False), ("U^^", False)]),
        ]
        self.run_tests(builder, tests)

    def test_prefixes_strictness_high(self):
        builder = regen.RegexBuilder(disjoiner_mode=regen.Strictness.HIGH)
        tests = [
            (["U"], [("U^", False), ("U", True)]),
            (["U", "^"], [("U^", True), ("U", False), ("U^^", False)]),
        ]
        self.run_tests(builder, tests)


class TestAspirates(RegexBuilderTester):
    def test_strictness_low(self):
        builder = regen.RegexBuilder(aspirate_mode=regen.Strictness.LOW)
        tests = [
            (["A"], [("A", True), ("'A", True)]),
            (["'", "A"], [("A", True), ("'A", True)]),
            (
                ["A", "D", "E"],
                [
                    ("ADE", True),
                    ("'ADE", True),
                    ("A'DE", True),
                    ("AD'E", True),
                    ("'A'DE", True),
                    ("'AD'E", True),
                    ("A'D'E", True),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["'", "A", "D", "E"],
                [
                    ("ADE", True),
                    ("'ADE", True),
                    ("A'DE", True),
                    ("AD'E", True),
                    ("'A'DE", True),
                    ("'AD'E", True),
                    ("A'D'E", True),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["'", "A", "'", "D", "E"],
                [
                    ("ADE", True),
                    ("'ADE", True),
                    ("A'DE", True),
                    ("AD'E", True),
                    ("'A'DE", True),
                    ("'AD'E", True),
                    ("A'D'E", True),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["'", "A", "'", "D", "'", "E"],
                [
                    ("ADE", True),
                    ("'ADE", True),
                    ("A'DE", True),
                    ("AD'E", True),
                    ("'A'DE", True),
                    ("'AD'E", True),
                    ("A'D'E", True),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["A", "'", "D", "'", "E"],
                [
                    ("ADE", True),
                    ("'ADE", True),
                    ("A'DE", True),
                    ("AD'E", True),
                    ("'A'DE", True),
                    ("'AD'E", True),
                    ("A'D'E", True),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["'", "A", "D", "'", "E"],
                [
                    ("ADE", True),
                    ("'ADE", True),
                    ("A'DE", True),
                    ("AD'E", True),
                    ("'A'DE", True),
                    ("'AD'E", True),
                    ("A'D'E", True),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["A", "'", "D", "E"],
                [
                    ("ADE", True),
                    ("'ADE", True),
                    ("A'DE", True),
                    ("AD'E", True),
                    ("'A'DE", True),
                    ("'AD'E", True),
                    ("A'D'E", True),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["A", "D", "'", "E"],
                [
                    ("ADE", True),
                    ("'ADE", True),
                    ("A'DE", True),
                    ("AD'E", True),
                    ("'A'DE", True),
                    ("'AD'E", True),
                    ("A'D'E", True),
                    ("'A'D'E", True),
                    ("AD''E", False),
                ],
            ),
        ]
        self.run_tests(builder, tests)

    def test_strictness_medium(self):
        builder = regen.RegexBuilder(aspirate_mode=regen.Strictness.MEDIUM)
        tests = [
            (["A"], [("A", True), ("'A", True)]),
            (["'", "A"], [("A", False), ("'A", True)]),
            (
                ["A", "D", "E"],
                [
                    ("ADE", True),
                    ("'ADE", True),
                    ("A'DE", True),
                    ("AD'E", True),
                    ("'A'DE", True),
                    ("'AD'E", True),
                    ("A'D'E", True),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["'", "A", "D", "E"],
                [
                    ("ADE", False),
                    ("'ADE", True),
                    ("A'DE", False),
                    ("AD'E", False),
                    ("'A'DE", True),
                    ("'AD'E", True),
                    ("A'D'E", False),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["'", "A", "'", "D", "E"],
                [
                    ("ADE", False),
                    ("'ADE", False),
                    ("A'DE", False),
                    ("AD'E", False),
                    ("'A'DE", True),
                    ("'AD'E", False),
                    ("A'D'E", False),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["'", "A", "'", "D", "'", "E"],
                [
                    ("ADE", False),
                    ("'ADE", False),
                    ("A'DE", False),
                    ("AD'E", False),
                    ("'A'DE", False),
                    ("'AD'E", False),
                    ("A'D'E", False),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["A", "'", "D", "'", "E"],
                [
                    ("ADE", False),
                    ("'ADE", False),
                    ("A'DE", False),
                    ("AD'E", False),
                    ("'A'DE", False),
                    ("'AD'E", False),
                    ("A'D'E", True),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["'", "A", "D", "'", "E"],
                [
                    ("ADE", False),
                    ("'ADE", False),
                    ("A'DE", False),
                    ("AD'E", False),
                    ("'A'DE", False),
                    ("'AD'E", True),
                    ("A'D'E", False),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["A", "'", "D", "E"],
                [
                    ("ADE", False),
                    ("'ADE", False),
                    ("A'DE", True),
                    ("AD'E", False),
                    ("'A'DE", True),
                    ("'AD'E", False),
                    ("A'D'E", True),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["A", "D", "'", "E"],
                [
                    ("ADE", False),
                    ("'ADE", False),
                    ("A'DE", False),
                    ("AD'E", True),
                    ("'A'DE", False),
                    ("'AD'E", True),
                    ("A'D'E", True),
                    ("'A'D'E", True),
                    ("AD''E", False),
                ],
            ),
        ]
        self.run_tests(builder, tests)

    def test_strictness_high(self):
        builder = regen.RegexBuilder(aspirate_mode=regen.Strictness.HIGH)
        tests = [
            (["A"], [("A", True), ("'A", False)]),
            (["'", "A"], [("A", False), ("'A", True)]),
            (
                ["A", "D", "E"],
                [
                    ("ADE", True),
                    ("'ADE", False),
                    ("A'DE", False),
                    ("AD'E", False),
                    ("'A'DE", False),
                    ("'AD'E", False),
                    ("A'D'E", False),
                    ("'A'D'E", False),
                ],
            ),
            (
                ["'", "A", "D", "E"],
                [
                    ("ADE", False),
                    ("'ADE", True),
                    ("A'DE", False),
                    ("AD'E", False),
                    ("'A'DE", False),
                    ("'AD'E", False),
                    ("A'D'E", False),
                    ("'A'D'E", False),
                ],
            ),
            (
                ["'", "A", "'", "D", "E"],
                [
                    ("ADE", False),
                    ("'ADE", False),
                    ("A'DE", False),
                    ("AD'E", False),
                    ("'A'DE", True),
                    ("'AD'E", False),
                    ("A'D'E", False),
                    ("'A'D'E", False),
                ],
            ),
            (
                ["'", "A", "'", "D", "'", "E"],
                [
                    ("ADE", False),
                    ("'ADE", False),
                    ("A'DE", False),
                    ("AD'E", False),
                    ("'A'DE", False),
                    ("'AD'E", False),
                    ("A'D'E", False),
                    ("'A'D'E", True),
                ],
            ),
            (
                ["A", "'", "D", "'", "E"],
                [
                    ("ADE", False),
                    ("'ADE", False),
                    ("A'DE", False),
                    ("AD'E", False),
                    ("'A'DE", False),
                    ("'AD'E", False),
                    ("A'D'E", True),
                    ("'A'D'E", False),
                ],
            ),
            (
                ["'", "A", "D", "'", "E"],
                [
                    ("ADE", False),
                    ("'ADE", False),
                    ("A'DE", False),
                    ("AD'E", False),
                    ("'A'DE", False),
                    ("'AD'E", True),
                    ("A'D'E", False),
                    ("'A'D'E", False),
                ],
            ),
            (
                ["A", "'", "D", "E"],
                [
                    ("ADE", False),
                    ("'ADE", False),
                    ("A'DE", True),
                    ("AD'E", False),
                    ("'A'DE", False),
                    ("'AD'E", False),
                    ("A'D'E", False),
                    ("'A'D'E", False),
                ],
            ),
            (
                ["A", "D", "'", "E"],
                [
                    ("ADE", False),
                    ("'ADE", False),
                    ("A'DE", False),
                    ("AD'E", True),
                    ("'A'DE", False),
                    ("'AD'E", False),
                    ("A'D'E", False),
                    ("'A'D'E", False),
                    ("AD''E", False),
                ],
            ),
        ]
        self.run_tests(builder, tests)

    def test_ing_strictness_low(self):
        builder = regen.RegexBuilder(aspirate_mode=regen.Strictness.LOW)
        tests = [
            (
                ["TH"],
                [
                    ("TH", True),
                    ("TH'", True),
                    ("TH''", True),
                    ("'TH'", True),
                    ("THE'", False),
                    ("TH'''", False),
                ],
            ),
            (
                ["TH", "'"],
                [
                    ("TH", True),
                    ("TH'", True),
                    ("TH''", True),
                    ("'TH'", True),
                    ("THE'", False),
                    ("TH'''", False),
                ],
            ),
            (
                ["TH", "'", "'"],
                [
                    ("TH", True),
                    ("TH'", True),
                    ("TH''", True),
                    ("'TH'", True),
                    ("THE'", False),
                    ("TH'''", False),
                ],
            ),
        ]
        self.run_tests(builder, tests)

    def test_ing_strictness_medium(self):
        builder = regen.RegexBuilder(aspirate_mode=regen.Strictness.MEDIUM)
        tests = [
            (
                ["TH"],
                [
                    ("TH", True),
                    ("TH'", True),
                    ("TH''", True),
                    ("'TH'", True),
                    ("THE'", False),
                    ("TH'''", False),
                ],
            ),
            (
                ["TH", "'"],
                [
                    ("TH", False),
                    ("TH'", True),
                    ("TH''", True),
                    ("'TH'", True),
                    ("THE'", False),
                    ("TH'''", False),
                ],
            ),
            (
                ["TH", "'", "'"],
                [
                    ("TH", False),
                    ("TH'", False),
                    ("TH''", True),
                    ("'TH'", False),
                    ("THE'", False),
                    ("TH'''", False),
                ],
            ),
        ]
        self.run_tests(builder, tests)

    def test_ing_strictness_high(self):
        builder = regen.RegexBuilder(aspirate_mode=regen.Strictness.HIGH)
        tests = [
            (
                ["TH"],
                [
                    ("TH", True),
                    ("TH'", False),
                    ("TH''", False),
                    ("'TH'", False),
                    ("THE'", False),
                    ("TH'''", False),
                ],
            ),
            (
                ["TH", "'"],
                [
                    ("TH", False),
                    ("TH'", True),
                    ("TH''", False),
                    ("'TH'", False),
                    ("THE'", False),
                    ("TH'''", False),
                ],
            ),
            (
                ["TH", "'", "'"],
                [
                    ("TH", False),
                    ("TH'", False),
                    ("TH''", True),
                    ("'TH'", False),
                    ("THE'", False),
                    ("TH'''", False),
                ],
            ),
        ]
        self.run_tests(builder, tests)

    def test_double_aspirate_strictness_low(self):
        builder = regen.RegexBuilder(aspirate_mode=regen.Strictness.LOW)
        tests = [
            (
                ["E", "D"],
                [("ED", True), ("'ED", True), ("''ED", True), ("'''ED", False)],
            ),
            (
                ["'", "E", "D"],
                [("ED", True), ("'ED", True), ("''ED", True), ("'''ED", False)],
            ),
            (
                ["'", "'", "E", "D"],
                [("ED", True), ("'ED", True), ("''ED", True), ("'''ED", False)],
            ),
        ]
        self.run_tests(builder, tests)

    def test_double_aspirate_strictness_medium(self):
        builder = regen.RegexBuilder(aspirate_mode=regen.Strictness.MEDIUM)
        tests = [
            (
                ["E", "D"],
                [("ED", True), ("'ED", True), ("''ED", True), ("'''ED", False)],
            ),
            (
                ["'", "E", "D"],
                [("ED", False), ("'ED", True), ("''ED", True), ("'''ED", False)],
            ),
            (
                ["'", "'", "E", "D"],
                [("ED", False), ("'ED", False), ("''ED", True), ("'''ED", False)],
            ),
        ]
        self.run_tests(builder, tests)

    def test_double_aspirate_strictness_high(self):
        builder = regen.RegexBuilder(aspirate_mode=regen.Strictness.HIGH)
        tests = [
            (
                ["E", "D"],
                [("ED", True), ("'ED", False), ("''ED", False), ("'''ED", False)],
            ),
            (
                ["'", "E", "D"],
                [("ED", False), ("'ED", True), ("''ED", False), ("'''ED", False)],
            ),
            (
                ["'", "'", "E", "D"],
                [("ED", False), ("'ED", False), ("''ED", True), ("'''ED", False)],
            ),
        ]
        self.run_tests(builder, tests)


class TestUncertaintyRegex(unittest.TestCase):
    def run_tests(self, uncertainty, tests):
        builder = regen.RegexBuilder()
        for test in tests:
            stroke = test[0]
            for text, expected in test[1]:
                regex = builder.make_uncertainty_regex(stroke, uncertainty)
                with self.subTest():
                    if expected:
                        self.assertTrue(re.fullmatch(regex, text))
                    else:
                        self.assertFalse(re.fullmatch(regex, text))

    # def test_uncertainty_zero(self):
    # # this should go under sim tests
    # for stroke in grammar.STROKES:
    # similars = similarities.get_similar(stroke, 0)
    # self.assertEqual(len(similars), 1)
    # self.assertIn(stroke, list(similars)[0])

    def test_uncertainty_zero(self):
        tests = [
            ("A", [("A", True), ("E", False), ("I", False)]),
            (
                "NT",
                [
                    ("NT", True),
                    ("ND", True),
                    ("TH", False),
                    ("MT", False),
                    ("MD", False),
                ],
            ),
            (
                "F",
                [
                    ("F", True),
                    ("V", False),
                    ("CH", False),
                    ("S", False),
                    ("Z", False),
                    ("SH", False),
                    ("P", False),
                    ("J", False),
                    ("X", False),
                    ("XS", False),
                    ("SS", False),
                    ("B", False),
                ],
            ),
        ]
        self.run_tests(0, tests)

    def test_uncertainty_one(self):
        tests = [
            ("A", [("A", True), ("E", True), ("I", True)]),
            (
                "NT",
                [("NT", True), ("ND", True), ("TH", True), ("MT", True), ("MD", True)],
            ),
            (
                "F",
                [
                    ("F", True),
                    ("V", True),
                    ("CH", True),
                    ("S", True),
                    ("Z", True),
                    ("SH", False),
                    ("P", False),
                    ("J", False),
                    ("X", False),
                    ("XS", False),
                    ("SS", False),
                    ("B", False),
                ],
            ),
        ]
        self.run_tests(1, tests)

    def test_uncertainty_two(self):
        tests = [
            ("A", [("A", True), ("E", True), ("I", True)]),
            (
                "NT",
                [("NT", True), ("ND", True), ("TH", True), ("MT", True), ("MD", True)],
            ),
            (
                "F",
                [
                    ("F", True),
                    ("V", True),
                    ("CH", True),
                    ("S", True),
                    ("Z", True),
                    ("SH", True),
                    ("P", True),
                    ("J", True),
                    ("X", True),
                    ("XS", True),
                    ("SS", True),
                    ("B", False),
                ],
            ),
        ]
        self.run_tests(2, tests)


class TestSearchMode(RegexBuilderTester):
    def test_match(self):
        builder = regen.RegexBuilder(search_mode=regen.SearchMode.MATCH)
        tests = [(["A", "B"], [("AB", True), ("ABU", False), ("DAB", False)])]
        self.run_tests(builder, tests)

    def test_start(self):
        builder = regen.RegexBuilder(search_mode=regen.SearchMode.START)
        tests = [(["A", "B"], [("AB", True), ("ABU", True), ("DAB", False)])]
        self.run_tests(builder, tests)

    def test_contain(self):
        builder = regen.RegexBuilder(search_mode=regen.SearchMode.CONTAIN)
        tests = [
            (["A", "B"], [("AB", True), ("ABU", True), ("DAB", True)]),
            (
                ["X"],
                [
                    ("ANEX Annex", True),
                    ("VEXN Vixen", True),
                    ("ZANTHUS Xanthous", False),
                    ("ZEBK Xebec", False),
                    ("ZEBK xebec", False),
                    ("ZEBK eXbec", False),
                    ("ZEBK ebec abXD", False),
                ],
            ),
        ]
        self.run_tests(builder, tests)


class TestFixFirst(RegexBuilderTester):
    def test_fix_first_off(self):
        builder = regen.RegexBuilder(fix_first=False, uncertainty=1)
        tests = [
            (
                ["A", "B", "D"],
                [
                    ("ABD", True),
                    ("EBD", True),
                    ("IBD", True),
                    ("APT", True),
                    ("EPDD", True),
                    ("IBDT", True),
                ],
            )
        ]
        self.run_tests(builder, tests)

    def test_fix_first_on(self):
        builder = regen.RegexBuilder(fix_first=True, uncertainty=1)
        tests = [
            (
                ["A", "B", "D"],
                [
                    ("ABD", True),
                    ("EBD", False),
                    ("IBD", False),
                    ("APT", True),
                    ("EPDD", False),
                    ("IBDT", False),
                ],
            )
        ]
        self.run_tests(builder, tests)


class TestStartingLetters(unittest.TestCase):
    def run_tests(self, builder, tests):
        for test in tests:
            interps = test[0]
            for stroke, expected in test[1]:
                letters = builder.get_starting_letters(interps)
                with self.subTest(interpretations=interps, stroke=stroke):
                    if expected:
                        self.assertIn(stroke, letters)
                    else:
                        self.assertNotIn(stroke, letters)

    def test_starting_letters_search_mode_contains(self):
        builder = regen.RegexBuilder(search_mode=regen.SearchMode.CONTAIN)
        tests = [[["A"]], [["B"]], [["K", "P"]], [["'", "I"]]]
        for test in tests:
            letters = builder.get_starting_letters(test)
            self.assertSetEqual(letters, grammar.HARD_CHARACTERS)

    def test_starting_letters_fix_first_uncertainty_one(self):
        builder = regen.RegexBuilder(
            search_mode=regen.SearchMode.MATCH, uncertainty=1, fix_first=True
        )
        tests = [
            ([["A"]], [("A", True), ("E", False), ("I", False)]),
            ([["'", "I"], [("I", True), ("E", False), ("A", False)]]),
            ([["NT"], ["DN"]], [("N", True), ("T", True), ("D", True), ("M", False)]),
        ]
        self.run_tests(builder, tests)

    def test_starting_letters_no_fix_first_uncertainty_one(self):
        builder = regen.RegexBuilder(
            search_mode=regen.SearchMode.MATCH, uncertainty=1, fix_first=False
        )
        tests = [
            ([["A"]], [("A", True), ("E", True), ("I", True)]),
            ([["'", "I"], [("I", True), ("E", False), ("A", True)]]),
            ([["NT"], ["DN"]], [("N", True), ("T", True), ("D", True), ("M", True)]),
        ]
        self.run_tests(builder, tests)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import logging
import unittest
from pathlib import Path

import pytest

from grascii.dictionary import Dictionary
from grascii.dictionary.build import (
    DEFAULT_PIPELINE,
    DictionaryBuilder,
    DictionaryOutputOptions,
)
from grascii.dictionary.common import (
    DictionaryAlreadyExists,
    DictionaryNotFound,
    get_dictionary_installed_name,
    get_dictionary_path_name,
)
from grascii.dictionary.install import install_dictionary
from grascii.dictionary.list import get_built_ins, get_installed
from grascii.dictionary.pipeline import create_grascii_check, create_spell_check
from grascii.dictionary.uninstall import uninstall_dictionary


class TestDictionaryBuildWarnings(unittest.TestCase):
    output_dir = Path("tests/dictionaries/outputs")

    def setUp(self):
        self.output_dir.mkdir(exist_ok=True)
        for entry in self.output_dir.iterdir():
            entry.unlink()

    def tearDown(self):
        for entry in self.output_dir.iterdir():
            entry.unlink()
        self.output_dir.rmdir()

    def test_uncertainty(self):
        builder = DictionaryBuilder(
            count_words=True,
        )
        summary = builder.build(
            infiles=[Path("tests/dictionaries/uncertainty.txt")],
            output=DictionaryOutputOptions(self.output_dir),
        )
        self.assertEqual(len(summary.warnings), 9)
        self.assertEqual(len(summary.errors), 0)
        for warning in summary.warnings:
            self.assertEqual(warning.level, logging.WARNING)
            self.assertRegex(warning.message, "Uncertainty")
        entry_count = sum(val for val in summary.entry_counts.values())
        self.assertEqual(entry_count, 9)

    def test_count(self):
        builder = DictionaryBuilder(
            count_words=True,
        )
        summary = builder.build(
            infiles=[Path("tests/dictionaries/count.txt")],
            output=DictionaryOutputOptions(self.output_dir),
        )
        self.assertEqual(len(summary.warnings), 10)
        self.assertEqual(len(summary.errors), 0)
        for warning in summary.warnings:
            self.assertEqual(warning.level, logging.WARNING)
            self.assertRegex(warning.message, "Too many")
        entry_count = sum(val for val in summary.entry_counts.values())
        self.assertEqual(entry_count, 12)

    def test_spelling(self):
        builder = DictionaryBuilder(
            pipeline=[
                *DEFAULT_PIPELINE,
                create_spell_check(
                    words_file=Path("tests/dictionaries/words.txt"),
                ),
            ],
            count_words=True,
        )
        summary = builder.build(
            infiles=[Path("tests/dictionaries/spell.txt")],
            output=DictionaryOutputOptions(self.output_dir),
        )
        self.assertEqual(len(summary.warnings), 5)
        self.assertEqual(len(summary.errors), 0)
        for warning in summary.warnings:
            self.assertEqual(warning.level, logging.WARNING)
            self.assertRegex(warning.message, "not in words")
        entry_count = sum(val for val in summary.entry_counts.values())
        self.assertEqual(entry_count, 9)

    def test_too_few_words(self):
        builder = DictionaryBuilder(
            count_words=True,
        )
        summary = builder.build(
            infiles=[Path("tests/dictionaries/too_few_words.txt")],
            output=DictionaryOutputOptions(self.output_dir),
        )
        self.assertEqual(len(summary.warnings), 0)
        self.assertEqual(len(summary.errors), 9)
        for error in summary.errors:
            self.assertEqual(error.level, logging.ERROR)
            self.assertRegex(error.message, "Too few")
        entry_count = sum(val for val in summary.entry_counts.values())
        self.assertEqual(entry_count, 0)

    def test_not_grascii(self):
        builder = DictionaryBuilder(
            pipeline=[*DEFAULT_PIPELINE, create_grascii_check()]
        )
        summary = builder.build(
            infiles=[Path("tests/dictionaries/not_grascii.txt")],
            output=DictionaryOutputOptions(self.output_dir),
        )
        self.assertEqual(len(summary.warnings), 0)
        self.assertEqual(len(summary.errors), 10)
        for error in summary.errors:
            self.assertEqual(error.level, logging.ERROR)
            self.assertRegex(error.message, "Failed to parse")
        entry_count = sum(val for val in summary.entry_counts.values())
        self.assertEqual(entry_count, 1)

    def test_comments_and_whitespace(self):
        builder = DictionaryBuilder()
        summary = builder.build(
            infiles=[Path("tests/dictionaries/comments.txt")],
            output=DictionaryOutputOptions(self.output_dir),
        )
        self.assertEqual(len(summary.warnings), 6)
        self.assertEqual(len(summary.errors), 0)
        for warning in summary.warnings:
            self.assertEqual(warning.level, logging.WARNING)
            self.assertRegex(warning.message, "Uncertainty")
        entry_count = sum(val for val in summary.entry_counts.values())
        self.assertEqual(entry_count, 7)


@pytest.mark.slow
class TestBuiltins(unittest.TestCase):
    def test_preanniversary(self):
        inputs = Path("dictionaries/builtins/preanniversary/").glob("*.txt")
        builder = DictionaryBuilder(
            count_words=True, pipeline=[*DEFAULT_PIPELINE, create_grascii_check()]
        )
        summary = builder.build(inputs, None)
        # Expecting one intentional error for "foyer"
        self.assertEqual(len(summary.errors), 1)


@pytest.fixture
def tmp_dict_path(tmp_path, monkeypatch):
    dict_path = tmp_path / "installed"
    dict_path.mkdir()
    monkeypatch.setattr("grascii.dictionary.INSTALLATION_DIR", dict_path)
    monkeypatch.setattr("grascii.dictionary.install.INSTALLATION_DIR", dict_path)
    monkeypatch.setattr("grascii.dictionary.uninstall.INSTALLATION_DIR", dict_path)
    monkeypatch.setattr("grascii.dictionary.list.INSTALLATION_DIR", dict_path)
    return dict_path


@pytest.fixture(scope="module")
def tmp_build_path(tmp_path_factory):
    build_path = tmp_path_factory.mktemp("search", numbered=False)
    builder = DictionaryBuilder()
    builder.build(
        infiles=Path("tests/dictionaries/search.txt"),
        output=DictionaryOutputOptions(build_path),
    )
    return build_path


class TestNewDictionary:
    def test_builtin(self):
        Dictionary.new(":preanniversary")

    def test_installed_does_not_exist(self):
        with pytest.raises(DictionaryNotFound):
            Dictionary.new(":unknown")

    def test_installed(self, tmp_dict_path, tmp_build_path):
        assert install_dictionary(tmp_build_path, tmp_dict_path) == ":search"
        Dictionary.new(":search")

    def test_path(self):
        Dictionary.new(Path("grascii/dictionary/preanniversary"))

    def test_string(self):
        Dictionary.new("grascii/dictionary/preanniversary")

    def test_path_does_not_exist(self):
        with pytest.raises(DictionaryNotFound):
            Dictionary.new(Path("unknown"))

    def test_string_does_not_exist(self):
        with pytest.raises(DictionaryNotFound):
            Dictionary.new("unknown")


class TestList:
    def test_list_no_installed(self, tmp_dict_path):
        assert len(get_installed()) == 0

    def test_list_builtins(self, tmp_dict_path):
        builtins = get_built_ins()
        assert len(builtins) == 1
        assert ":preanniversary" in builtins


class TestInstall:
    def test_install(self, tmp_dict_path, tmp_build_path):
        assert len(get_installed()) == 0
        assert install_dictionary(tmp_build_path, tmp_dict_path) == ":search"
        assert len(get_installed()) == 1
        assert ":search" in get_installed()

    def test_uninstall(self, tmp_dict_path, tmp_build_path):
        assert len(get_installed()) == 0
        assert install_dictionary(tmp_build_path, tmp_dict_path) == ":search"
        assert len(get_installed()) == 1
        uninstall_dictionary("search", tmp_dict_path)
        assert len(get_installed()) == 0

    def test_uninstall_nonexistent(self, tmp_dict_path):
        assert len(get_installed()) == 0
        with pytest.raises(DictionaryNotFound):
            uninstall_dictionary("search", tmp_dict_path)

    def test_install_name(self, tmp_dict_path, tmp_build_path):
        assert len(get_installed()) == 0
        assert install_dictionary(tmp_build_path, tmp_dict_path, name="two") == ":two"
        assert len(get_installed()) == 1
        assert ":two" in get_installed()

    def test_install_twice(self, tmp_dict_path, tmp_build_path):
        assert len(get_installed()) == 0
        assert install_dictionary(tmp_build_path, tmp_dict_path) == ":search"
        assert len(get_installed()) == 1
        with pytest.raises(DictionaryAlreadyExists):
            install_dictionary(tmp_build_path, tmp_dict_path)

    def test_install_force(self, tmp_dict_path, tmp_build_path):
        assert len(get_installed()) == 0
        assert install_dictionary(tmp_build_path, tmp_dict_path) == ":search"
        assert len(get_installed()) == 1
        with pytest.raises(DictionaryAlreadyExists):
            install_dictionary(tmp_build_path, tmp_dict_path)
        assert (
            install_dictionary(tmp_build_path, tmp_dict_path, force=True) == ":search"
        )


class TestCommon:
    def test_installed_name(self):
        assert get_dictionary_installed_name(":preanniversary") == ":preanniversary"
        assert get_dictionary_installed_name("preanniversary") == ":preanniversary"
        with pytest.raises(ValueError):
            get_dictionary_installed_name("")
        with pytest.raises(ValueError):
            get_dictionary_installed_name(":")

    def test_path_name(self):
        assert get_dictionary_path_name(":preanniversary") == "preanniversary"
        assert get_dictionary_path_name("preanniversary") == "preanniversary"
        with pytest.raises(ValueError):
            get_dictionary_path_name("")
        with pytest.raises(ValueError):
            get_dictionary_path_name(":")



from __future__ import annotations

import unittest
from pathlib import Path
from shutil import rmtree

from grascii.dictionary import DictionaryNotFound
from grascii.dictionary.build import DictionaryBuilder, DictionaryOutputOptions
from grascii.parser import InvalidGrascii
from grascii.searchers import GrasciiSearcher, RegexSearcher, ReverseSearcher

output_dir = "tests/dictionaries/tosearch"
sorted_output_dir = "test/dictionaries/sorted"


def build_dictionary(src, dest):
    rmtree(dest, ignore_errors=True)
    infiles = [Path(src)]
    builder = DictionaryBuilder()
    builder.build(infiles=infiles, output=DictionaryOutputOptions(dest))


def setUpModule():
    build_dictionary("tests/dictionaries/search.txt", output_dir)
    build_dictionary("tests/dictionaries/sort.txt", sorted_output_dir)


def tearDownModule():
    rmtree(output_dir, ignore_errors=True)
    rmtree(sorted_output_dir, ignore_errors=True)


class TestGrasciiSearcher(unittest.TestCase):
    def test_uncertainty(self):
        searcher = GrasciiSearcher(dictionaries=[output_dir])
        uncertainty_zero_result_count = len(
            searcher.sorted_search(grascii="FTH", uncertainty=0)
        )
        uncertainty_one_result_count = len(
            searcher.sorted_search(grascii="FTH", uncertainty=1)
        )
        uncertainty_two_result_count = len(
            searcher.sorted_search(grascii="FTH", uncertainty=2)
        )
        self.assertGreater(uncertainty_one_result_count, uncertainty_zero_result_count)
        self.assertGreater(uncertainty_two_result_count, uncertainty_one_result_count)

    def test_fix_first(self):
        searcher = GrasciiSearcher(dictionaries=[output_dir])
        fix_first_off_count = len(
            searcher.sorted_search(grascii="FTH", fix_first=False, uncertainty=1)
        )
        fix_first_on_count = len(
            searcher.sorted_search(grascii="FTH", fix_first=True, uncertainty=1)
        )
        self.assertGreater(fix_first_off_count, fix_first_on_count)

    def test_interpretation(self):
        searcher = GrasciiSearcher(dictionaries=[output_dir])
        best_interpretation_count = len(
            searcher.sorted_search(grascii="SSTN", interpretation="best", uncertainty=2)
        )
        all_interpretation_count = len(
            searcher.sorted_search(grascii="SSTN", interpretation="all", uncertainty=2)
        )
        self.assertGreater(all_interpretation_count, best_interpretation_count)

    def test_search_mode(self):
        searcher = GrasciiSearcher(dictionaries=[output_dir])
        match_count = len(searcher.sorted_search(grascii="ABT", search_mode="match"))
        start_count = len(searcher.sorted_search(grascii="ABT", search_mode="start"))
        contain_count = len(
            searcher.sorted_search(grascii="ABT", search_mode="contain")
        )
        self.assertGreater(start_count, match_count)
        self.assertGreater(contain_count, start_count)

    def test_invalid_grascii(self):
        searcher = GrasciiSearcher(dictionaries=[output_dir])
        with self.assertRaises(InvalidGrascii):
            searcher.sorted_search(grascii="RAC")
        with self.assertRaises(InvalidGrascii):
            searcher.sorted_search(grascii="WAY")
        with self.assertRaises(InvalidGrascii):
            searcher.sorted_search(grascii="YES")

    def test_dictionary_not_found(self):
        with self.assertRaises(DictionaryNotFound):
            GrasciiSearcher(dictionaries=[":should-not-exist"])


class TestSortedGrasciiSearches(unittest.TestCase):
    def test_shorter_grascii_first(self):
        searcher = GrasciiSearcher(dictionaries=[sorted_output_dir])
        results = searcher.sorted_search(grascii="ABS", search_mode="start")
        self.assertListEqual(
            [r.entry.grascii for r in results],
            [
                "ABS",
                "ABSS",
                "ABSOL",
                "ABSTMUS",
            ],
        )

    def test_smaller_offset_match_first(self):
        searcher = GrasciiSearcher(dictionaries=[sorted_output_dir])
        results = searcher.sorted_search(grascii="MAS", search_mode="contain")
        self.assertListEqual(
            [r.entry.grascii for r in results],
            [
                "MAS",
                "AMAS",
                "PAMAS",
                "TASKMAS",
            ],
        )

    def test_better_matches_first(self):
        searcher = GrasciiSearcher(dictionaries=[sorted_output_dir])
        results = searcher.sorted_search(
            grascii="RELES", search_mode="contain", uncertainty=1
        )
        self.assertListEqual(
            [r.entry.grascii for r in results],
            [
                "RELES",
                "RELESB",
                "RELEZSH",
                "RELEF",
                "RELAPS",
                "RELAXSH",
                "RAREF",
            ],
        )

    def test_major_annotation_differences_appear_later(self):
        searcher = GrasciiSearcher(dictionaries=[sorted_output_dir])
        results = searcher.sorted_search(grascii="TAD", search_mode="contain")
        self.assertListEqual(
            [r.entry.grascii for r in results],
            [
                "STAD",
                "STADEM",
                "SETADL",
                "TA~DE|",
                "TA~DEN",
                "STA~D",
                "KUSTA~D",
                "DASTA~DE",
            ],
        )

    def test_minor_annotation_differences_special_handling(self):
        searcher = GrasciiSearcher(dictionaries=[sorted_output_dir])
        results = searcher.sorted_search(grascii="PO-E", search_mode="contain")
        self.assertListEqual(
            [r.entry.grascii for r in results],
            [
                "POES",
                "POESN",
                "POENANT",
                "POENANSE",
                "PO,EM",
                "PO,ESE",
                "PO,ETRE",
                "SPOEL",
                "SPOELR",
            ],
        )

    def test_exact_matches_appear_on_top(self):
        searcher = GrasciiSearcher(dictionaries=[sorted_output_dir])
        results = searcher.sorted_search(grascii="PO,E", search_mode="contain")
        self.assertListEqual(
            [r.entry.grascii for r in results],
            [
                "PO,EM",
                "PO,ESE",
                "PO,ETRE",
                "POES",
                "POESN",
                "POENANT",
                "POENANSE",
                "SPOEL",
                "SPOELR",
            ],
        )


class TestReverseSearcher(unittest.TestCase):
    def test_results(self):
        searcher = ReverseSearcher(dictionaries=[output_dir])
        result_count = len(searcher.sorted_search(reverse="habit"))
        self.assertEqual(result_count, 8)

    def test_no_grascii_match(self):
        searcher = ReverseSearcher(dictionaries=[output_dir])
        result_count = len(searcher.sorted_search(reverse="'ABT"))
        self.assertEqual(result_count, 0)

    def test_dictionary_not_found(self):
        with self.assertRaises(DictionaryNotFound):
            ReverseSearcher(dictionaries=["does-not-exist"])

    def test_regex_escape(self):
        searcher = ReverseSearcher(dictionaries=[output_dir])
        result_count = len(searcher.sorted_search(reverse="ag+.*"))
        self.assertEqual(result_count, 0)


class TestRegexSeacrher(unittest.TestCase):
    def test_dictionary_not_found(self):
        with self.assertRaises(DictionaryNotFound):
            RegexSearcher(dictionaries=[":preanniversary", ":cannot-exist"])



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Callable, Tuple

PipelineFunc = Callable[[str, str, logging.LoggerAdapter], Tuple[str, str]]
"""A function for dictionary builds that checks and/or
transforms a Grascii string and its translation in a dictionary entry.

The first parameter is a Grascii string, the second parameter is its translation,
and the third parameter is a logger. The function must return a Grascii string
and a translation or raise a ``CancelPipeline`` exception.
"""


class CancelPipeline(Exception):
    """An Exception that can be thrown from a pipeline function to cancel the
    entire pipeline for a given entry."""

    pass


def create_grascii_check(ignore_case: bool = True) -> PipelineFunc:
    """Create a pipeline function that validates the Grascii string.

    :param ignore_case: Whether to ignore the case of the Grascii string. If
        ``False``, the Grascii string must be uppercase.
    :type ignore_case: bool
    """
    from grascii.parser import GrasciiValidator

    # Disable cache for now
    # It could be enabled, but we have to be careful about clearing the
    # cache after grammar changes
    validator = GrasciiValidator(ignore_case=ignore_case)

    def check_grascii(grascii: str, translation: str, logger: logging.LoggerAdapter):
        if not validator.validate(grascii):
            logger.error(f"Failed to parse {grascii}")
            raise CancelPipeline
        return grascii, translation

    return check_grascii


def create_spell_check(words_file: os.PathLike) -> PipelineFunc:
    """Create a pipeline function that checks the words in a translation using
    the words from a file.

    :param words_file: A path to a line-delimited words file.
    :type words_file: os.PathLike
    """

    words = set()
    with Path(words_file).open("r") as f:
        words |= set(line.strip().upper() for line in f)

    def spell_check(grascii: str, translation: str, logger: logging.LoggerAdapter):
        translation_words = [w.upper() for w in translation.split()]
        for translation_word in translation_words:
            if translation_word not in words:
                logger.warning(f"{translation_word} not in words file")
        return grascii, translation

    return spell_check


def remove_boundaries(grascii: str, translation: str, logger: logging.LoggerAdapter):
    """A pipeline function that removes boundary characters from the Grascii string."""
    return grascii.replace("-", ""), translation


def standardize_case(grascii: str, translation: str, logger: logging.LoggerAdapter):
    """A pipeline function that uppercases the Grascii string and capitalizes
    each word in the translation.
    """
    capitalized = " ".join(w.capitalize() for w in translation.split())
    return grascii.upper(), capitalized
